import json
from dataclasses import dataclass
from enum import Enum, auto
from io import TextIOWrapper
from pathlib import Path
from textwrap import dedent
from typing import List

import click
import yaml

from latch_cli.constants import latch_constants
from latch_cli.utils import WorkflowType
from latch_cli.workflow_config import LatchWorkflowConfig, create_and_write_config


class DockerCmdBlockOrder(str, Enum):
    """Put a command block before or after the primary COPY command."""

    precopy = auto()
    postcopy = auto()


@dataclass(frozen=True)
class DockerCmdBlock:
    comment: str
    commands: List[str]
    order: DockerCmdBlockOrder

    def write_block(self, f: TextIOWrapper):
        f.write(f"# {self.comment}\n")
        f.write("\n".join(self.commands) + "\n\n")


def get_prologue(
    config: LatchWorkflowConfig, wf_type: WorkflowType = WorkflowType.latchbiosdk
) -> List[str]:
    if wf_type == WorkflowType.snakemake:
        library_name = '"latch[snakemake]"'
    else:
        library_name = "latch"
    return [
        "# DO NOT CHANGE",
        f"from {config.base_image}",
        "",
        "workdir /tmp/docker-build/work/",
        "",
        dedent(r"""
        shell [ \
            "/usr/bin/env", "bash", \
            "-o", "errexit", \
            "-o", "pipefail", \
            "-o", "nounset", \
            "-o", "verbose", \
            "-o", "errtrace", \
            "-O", "inherit_errexit", \
            "-O", "shift_verbose", \
            "-c" \
        ]
        """).strip(),
        "env TZ='Etc/UTC'",
        "env LANG='en_US.UTF-8'",
        "",
        "arg DEBIAN_FRONTEND=noninteractive",
        "",
        "# Latch SDK",
        "# DO NOT REMOVE",
        f"run pip install {library_name}=={config.latch_version}",
        "run mkdir /opt/latch",
    ]


def get_epilogue(wf_type: WorkflowType = WorkflowType.latchbiosdk) -> List[str]:
    cmds: list[str] = []

    if wf_type == WorkflowType.snakemake:
        cmds += [
            "",
            "# Latch snakemake workflow entrypoint",
            "# DO NOT CHANGE",
            "copy .latch/snakemake_jit_entrypoint.py /root/snakemake_jit_entrypoint.py",
        ]

    cmds += [
        "",
        "# Latch workflow registration metadata",
        "# DO NOT CHANGE",
        "arg tag",
        "# DO NOT CHANGE",
        "env FLYTE_INTERNAL_IMAGE $tag",
        "",
        "workdir /root",
    ]

    return cmds


def infer_commands(pkg_root: Path) -> List[DockerCmdBlock]:
    commands: List[DockerCmdBlock] = []

    if (pkg_root / "system-requirements.txt").exists():
        click.echo(
            " ".join([
                click.style(f"system-requirements.txt:", bold=True),
                "System dependencies installation phase",
            ])
        )

        commands.append(
            DockerCmdBlock(
                comment="Install system dependencies",
                commands=[
                    "copy system-requirements.txt /opt/latch/system-requirements.txt",
                    dedent(r"""
                            run apt-get update --yes && \
                                xargs apt-get install --yes \
                                    < /opt/latch/system-requirements.txt
                            """).strip(),
                ],
                order=DockerCmdBlockOrder.precopy,
            )
        )

    if (pkg_root / "environment.R").exists():
        click.echo(
            " ".join([
                click.style(f"environment.R:", bold=True),
                "R dependencies installation phase",
            ])
        )

        # todo(maximsmol): allow specifying R version
        # todo(maximsmol): somehow promote using pak
        commands += [
            DockerCmdBlock(
                comment="Install rig the R installation manager",
                commands=[
                    dedent(r"""
                            run \
                                curl \
                                    --location \
                                    --fail \
                                    --remote-name \
                                    https://github.com/r-lib/rig/releases/download/latest/rig-linux-latest.tar.gz && \
                                tar \
                                    --extract \
                                    --gunzip \
                                    --file rig-linux-latest.tar.gz \
                                    --directory /usr/local/ && \
                                rm rig-linux-latest.tar.gz
                            """).strip(),
                ],
                order=DockerCmdBlockOrder.precopy,
            ),
            DockerCmdBlock(
                comment="Install R",
                commands=[
                    "run rig add release # Change to any R version",
                ],
                order=DockerCmdBlockOrder.precopy,
            ),
            DockerCmdBlock(
                comment="Install R dependencies",
                commands=[
                    "copy environment.R /opt/latch/environment.R",
                    "run Rscript /opt/latch/environment.R",
                ],
                order=DockerCmdBlockOrder.precopy,
            ),
        ]

    conda_env_p = pkg_root / "environment.yml"
    if not conda_env_p.exists():
        conda_env_p = conda_env_p.with_suffix(".yaml")

    if conda_env_p.exists():
        click.echo(
            " ".join([
                click.style(f"{conda_env_p.name}:", bold=True),
                "Conda dependencies installation phase",
            ])
        )

        with conda_env_p.open("rb") as f:
            conda_env = yaml.safe_load(f)

        if "name" in conda_env:
            env_name = conda_env["name"]
        else:
            env_name = "workflow"

        # todo(maximsmol): install `curl` and other build deps ahead of time once (or in base image)
        commands += [
            DockerCmdBlock(
                comment="Install Mambaforge",
                commands=[
                    dedent(r"""
                            run apt-get update --yes && \
                                apt-get install --yes curl && \
                                curl \
                                    --location \
                                    --fail \
                                    --remote-name \
                                    https://github.com/conda-forge/miniforge/releases/latest/download/Mambaforge-Linux-x86_64.sh && \
                                `# Docs for -b and -p flags: https://docs.anaconda.com/anaconda/install/silent-mode/#linux-macos` \
                                bash Mambaforge-Linux-x86_64.sh -b -p /opt/conda -u && \
                                rm Mambaforge-Linux-x86_64.sh
                            """).strip(),
                ],
                order=DockerCmdBlockOrder.precopy,
            ),
            DockerCmdBlock(
                comment="Set conda PATH",
                commands=[
                    "env PATH=/opt/conda/bin:$PATH",
                    "RUN conda config --set auto_activate_base false",
                ],
                order=DockerCmdBlockOrder.precopy,
            ),
            DockerCmdBlock(
                comment="Build conda environment",
                commands=[
                    f"copy {conda_env_p.name} /opt/latch/environment.yaml",
                    dedent(rf"""
                            run mamba env create \
                                --file /opt/latch/environment.yaml \
                                --name {env_name}
                            """).strip(),
                    f"env PATH=/opt/conda/envs/{env_name}/bin:$PATH",
                ],
                order=DockerCmdBlockOrder.precopy,
            ),
        ]

    has_setup_py = (pkg_root / "setup.py").exists()

    has_buildable_pyproject = False
    try:
        with (pkg_root / "pyproject.toml").open("r") as f:
            for line in f:
                if not line.startswith("[build-system]"):
                    continue

                has_buildable_pyproject = True
                break
    except FileNotFoundError:
        ...

    # from https://peps.python.org/pep-0518/ and https://peps.python.org/pep-0621/
    if has_setup_py or has_buildable_pyproject:
        cause = "setup.py" if has_setup_py else "pyproject.toml"
        click.echo(
            " ".join([
                click.style(f"{cause}:", bold=True),
                "Python package installation phase",
            ])
        )

        print()
        commands.append(
            DockerCmdBlock(
                comment="Install python package",
                commands=["run pip install /root/"],
                order=DockerCmdBlockOrder.postcopy,
            )
        )

    if (pkg_root / "requirements.txt").exists():
        click.echo(
            " ".join([
                click.style("requirements.txt:", bold=True),
                "Python pip dependencies installation phase",
            ])
        )
        commands.append(
            DockerCmdBlock(
                comment="Install pip dependencies from `requirements.txt`",
                commands=[
                    "copy requirements.txt /opt/latch/requirements.txt",
                    "run pip install --requirement /opt/latch/requirements.txt",
                ],
                order=DockerCmdBlockOrder.precopy,
            )
        )

    if (pkg_root / ".env").exists():
        click.echo(
            " ".join([click.style(".env:", bold=True), "Environment variable setup"])
        )
        envs: list[str] = []
        for line in (pkg_root / ".env").read_text().splitlines():
            line = line.strip()

            if line == "":
                continue
            if line.startswith("#"):
                continue

            envs.append(f"env {line}")

        commands.append(
            DockerCmdBlock(
                comment="Set environment variables",
                commands=envs,
                order=DockerCmdBlockOrder.precopy,
            )
        )

    return commands


def generate_dockerfile(
    pkg_root: Path, outfile: Path, *, wf_type: WorkflowType = WorkflowType.latchbiosdk
) -> None:
    """Generate a best effort Dockerfile from files in the workflow directory.

    Args:
        pkg_root: A path to a workflow directory.
        outfile: The path to write the generated Dockerfile.
        wf_type: The type of workflow (eg. snakemake) the Dockerfile is for

    Example:

        >>> generate_dockerfile(Path("test-workflow"), Path("test-workflow/Dockerfile"))
            # The resulting file structure will look like
            #   test-workflow
            #   ├── Dockerfile
            #   └── ...
    """

    click.secho("Generating Dockerfile", bold=True)
    try:
        with (pkg_root / latch_constants.pkg_config).open("r") as f:
            config: LatchWorkflowConfig = LatchWorkflowConfig(**json.load(f))
    except FileNotFoundError:
        click.secho("Creating a default configuration file")

        create_and_write_config(pkg_root)
        with (pkg_root / latch_constants.pkg_config).open("r") as f:
            config = LatchWorkflowConfig(**json.load(f))

    click.echo(
        " ".join([
            click.style("Base image:", fg="bright_blue"),
            config.base_image,
        ])
    )
    click.echo(
        " ".join([
            click.style("Latch SDK version:", fg="bright_blue"),
            config.latch_version,
        ])
    )
    click.echo()

    with outfile.open("w") as f:
        f.write("\n".join(get_prologue(config, wf_type)) + "\n\n")

        commands = infer_commands(pkg_root)
        if len(commands) > 0:
            click.echo()

        for block in commands:
            if block.order != DockerCmdBlockOrder.precopy:
                continue

            block.write_block(f)

        f.write("# Copy workflow data (use .dockerignore to skip files)\n")
        f.write("copy . /root/\n\n")

        for block in commands:
            if block.order != DockerCmdBlockOrder.postcopy:
                continue

            block.write_block(f)

        f.write("\n".join(get_epilogue(wf_type)) + "\n")


def get_default_dockerfile(pkg_root: Path, *, wf_type: WorkflowType):
    default_dockerfile = pkg_root / "Dockerfile"

    if not default_dockerfile.exists():
        default_dockerfile = pkg_root / ".latch" / "Dockerfile"
        generate_dockerfile(pkg_root, default_dockerfile, wf_type=wf_type)

    return default_dockerfile



from latch.types.directory import LatchDir, LatchOutputDir
from latch.types.file import LatchFile, LatchOutputFile
from latch.types.glob import file_glob
from latch.types.metadata import (
    DockerMetadata,
    Fork,
    ForkBranch,
    LatchAppearanceType,
    LatchAuthor,
    LatchMetadata,
    LatchParameter,
    LatchRule,
    Params,
    Section,
    Spoiler,
    Text,
)



"""
Minimal template workflow to show how to run workflows that use conda in Latch

For a more comprehensive template, see the assemble_and_sort workflow
For examples on how to use conda in Latch, see https://docs.latch.bio/examples/workflows_examples.html
"""

from wf.conda_task import conda_task

from latch.resources.tasks import small_task
from latch.resources.workflow import workflow
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile
from latch.types.metadata import LatchAuthor, LatchMetadata, LatchParameter

"""Minimal metadata object - fill in fields with your own values"""
metadata = LatchMetadata(
    display_name="CHANGE ME",
    documentation="CHANGE ME",
    author=LatchAuthor(
        name="CHANGE ME",
        email="CHANGE ME",
        github="CHANGE ME",
    ),
    repository="CHANGE ME",
    license="CHANGE ME",
    parameters={
        "input_file": LatchParameter(
            display_name="Input File",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
        "output_directory": LatchParameter(
            display_name="Output Directory",
            batch_table_column=True,  # Show this parameter in batched mode.
        ),
    },
    tags=[],
)


# change the name of this function to something more descriptive
@workflow(metadata)
def conda_workflow(
    input_file: LatchFile, output_directory: LatchOutputDir
) -> LatchFile:
    return conda_task(input_file=input_file, output_directory=output_directory)



import json as _json
import time
from http import HTTPStatus
from http.client import HTTPException, HTTPResponse, HTTPSConnection
from typing import Any, Dict, Optional
from urllib.parse import urlparse, urlunparse


class TinyResponse:
    def __init__(self, resp: HTTPResponse, url: str, *, stream: bool = False) -> None:
        self._resp = resp
        self._content = None
        self._url = url

        self._stream = stream
        if not self._stream:
            self._content = self._resp.read()

    @property
    def headers(self):
        return self._resp.headers

    @property
    def status_code(self):
        return self._resp.status

    def json(self):
        return _json.loads(self.content)

    @property
    def content(self):
        if self._content is None:
            self._content = self._resp.read()
        return self._content

    def iter_content(self, chunk_size: Optional[int] = 1):
        while True:
            if chunk_size is None:
                x = self._resp.read1()
            else:
                x = self._resp.read(chunk_size)

            if len(x) == 0:
                yield x
                break

            yield x

    @property
    def url(self):
        return self._url

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            err_type = "Client" if self.status_code < 500 else "Server"

            status_enum = HTTPStatus(self.status_code)
            raise HTTPException(
                f"{self.status_code} {err_type} error: {status_enum.phrase} {self.url}"
            )

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if self._stream:
            self._resp.close()


def _req(
    method: str,
    url: str,
    *,
    headers: Dict[str, str] = {},
    data: Optional[bytes] = None,
    json: Optional[Any] = None,
    stream: bool = False,
):
    parts = urlparse(url)
    if parts.hostname is None:
        raise ValueError(f"could not extract hostname from {url}")

    body = None
    if data is not None:
        body = data

    if json is not None:
        body = _json.dumps(json)
        headers["Content-Type"] = "application/json"

    conn = HTTPSConnection(
        parts.hostname, parts.port if parts.port is not None else 443
    )
    conn.request(
        method,
        urlunparse(parts._replace(scheme="", netloc="")),
        headers=headers,
        body=body,
    )
    resp = conn.getresponse()

    return TinyResponse(resp, url, stream=stream)


def request(
    method: str,
    url: str,
    *,
    headers: Optional[Dict[str, str]] = {},
    data: Optional[Any] = None,
    json: Optional[bytes] = None,
    stream: bool = False,
    num_retries: int = 3,
) -> TinyResponse:
    """
    Send HTTP request. Retry on 500s or ConnectionErrors.
    Implements exponential backoff between retries.
    """
    err = None
    res = None

    attempt = 0
    while attempt < num_retries:
        res = None
        attempt += 1
        try:
            res = _req(
                method, url, headers=headers, data=data, json=json, stream=stream
            )
            if res.status_code < 500:
                return res
        except ConnectionError as e:
            err = e

        if attempt < num_retries:
            # todo(rahul): tune the sleep interval based on the startup time of the server
            # todo(rahul): change sleep interval based on which service we are calling
            time.sleep(2**attempt * 5)

    if res is None:
        raise err
    return res


def get(
    url: str,
    *,
    headers: Dict[str, str] = {},
    data: Optional[bytes] = None,
    json: Optional[Any] = None,
    stream: bool = False,
    num_retries: int = 3,
):
    assert data is None or json is None, "At most one of `data` and `json` can be set"

    return request(
        "GET",
        url,
        headers=headers,
        data=data,
        json=json,
        stream=stream,
        num_retries=num_retries,
    )


def put(
    url: str,
    *,
    headers: Dict[str, str] = {},
    data: Optional[bytes] = None,
    json: Optional[Any] = None,
    stream: bool = False,
    num_retries: int = 3,
):
    assert data is None or json is None, "At most one of `data` and `json` can be set"

    return request(
        "PUT",
        url,
        headers=headers,
        data=data,
        json=json,
        stream=stream,
        num_retries=num_retries,
    )


def post(
    url: str,
    *,
    headers: Dict[str, str] = {},
    data: Optional[bytes] = None,
    json: Optional[Any] = None,
    stream: bool = False,
    num_retries: int = 3,
):
    assert data is None or json is None, "At most one of `data` and `json` can be set"

    return request(
        "POST",
        url,
        headers=headers,
        data=data,
        json=json,
        stream=stream,
        num_retries=num_retries,
    )



from typing import Dict, List, Literal, TypedDict, Union

from typing_extensions import NotRequired, TypeAlias

from latch.types.json import JsonObject


class PrimitiveTypeBasic(TypedDict):
    primitive: Union[
        Literal["string"],
        Literal["integer"],
        Literal["number"],
        Literal["datetime"],
        Literal["blob"],
        Literal["date"],
        Literal["null"],
        Literal["boolean"],
    ]
    metadata: NotRequired[JsonObject]


class PrimitiveTypeLink(TypedDict):
    primitive: Literal["link"]
    experimentId: str
    metadata: NotRequired[JsonObject]


class PrimitiveTypeEnum(TypedDict):
    primitive: Literal["enum"]
    members: List[str]
    metadata: NotRequired[JsonObject]


PrimitiveType: TypeAlias = Union[
    PrimitiveTypeBasic, PrimitiveTypeLink, PrimitiveTypeEnum
]

PrimitiveTypeName = Union[
    Literal["string"],
    Literal["integer"],
    Literal["number"],
    Literal["datetime"],
    Literal["blob"],
    Literal["date"],
    Literal["null"],
    Literal["boolean"],
    Literal["link"],
    Literal["enum"],
]


class UnionType(TypedDict):
    union: Dict[str, "RegistryType"]


class ArrayType(TypedDict):
    array: "RegistryType"


RegistryType: TypeAlias = Union[PrimitiveType, UnionType, ArrayType]


class DBType(TypedDict):
    type: RegistryType
    allowEmpty: bool



import re
from pathlib import Path
from textwrap import dedent
from urllib.parse import urlparse

import click
from latch_sdk_config.user import user_config

from latch_cli.utils import get_auth_header

latch_url_regex = re.compile(r"^(latch)?://")


def is_remote_path(path: str) -> bool:
    return latch_url_regex.match(path) is not None


scheme = re.compile(
    r"""
    ^(
        (latch://) |
        (file://) |
        (?P<implicit_url>://) |
        (?P<absolute_path>/) |
        (?P<relative_path>[^/])
    )
    """,
    re.VERBOSE,
)
domain = re.compile(
    r"""
    (?P<account_relative>
        ^$| # empty
        (?P<shared>^shared$)
    ) |
    (
        ^
        (
            (shared\.\d+\.account) |
            (\d+\.account) |
            ([^/]+\.mount) |
            (\d+\.node)
        )
        $
    )
    """,
    re.VERBOSE,
)


# scheme inference rules:
#   ://domain/a/b/c => latch://domain/a/b/c
#   /a/b/c => file:///a/b/c
#   a/b/c => file://${pwd}/a/b/c
#
# if assume_remote = True (i.e. in the ls case):
#   /a/b/c => latch:///a/b/c
#   a/b/c => latch:///a/b/c
def append_scheme(path: str, *, assume_remote: bool = False) -> str:
    match = scheme.match(path)
    if match is None:
        click.secho(
            dedent(f"""
            `{path}` is not in a valid format.

            See https://docs.latch.bio/basics/latch_urls.html for a list of valid remote path formats.
            """).strip("\n"),
            fg="red",
        )
        raise click.exceptions.Exit(1)

    if match["implicit_url"] is not None:
        path = f"latch{path}"
    elif match["absolute_path"] is not None:
        path = f"latch://{path}" if assume_remote else f"file://{path}"
    elif match["relative_path"] is not None:
        path = f"latch:///{path}" if assume_remote else f"file://{Path.cwd()}/{path}"

    return path


# domain inference rules:
#   latch:///a/b/c => latch://xxx.account/a/b/c
#   latch://shared/a/b/c => latch://shared.xxx.account/a/b/c
#   latch://any_other_domain/a/b/c => unchanged
def append_domain(path: str) -> str:
    workspace = user_config.workspace_id

    parsed = urlparse(path)
    dom = parsed.netloc

    if dom == "" and workspace != "":
        dom = f"{workspace}.account"

    match = domain.match(dom)
    if match is None:
        click.secho(
            dedent(f"""
            `{dom}` is not a valid path domain.

            See https://docs.latch.bio/basics/latch_urls.html for a list of valid domains.
            """).strip("\n"),
            fg="red",
        )
        raise click.exceptions.Exit(1)

    if match["shared"] is not None and workspace != "":
        dom = f"shared.{workspace}.account"

    return parsed._replace(netloc=dom).geturl()


strip_domain_expr = re.compile(r"^(latch)?://[^/]*/")


def strip_domain(path: str) -> str:
    return strip_domain_expr.sub("", path, 1)


def is_account_relative(path: str) -> bool:
    parsed = urlparse(path)
    dom = parsed.netloc

    match = domain.match(dom)
    if match is None:
        click.secho(
            dedent(f"""
            `{dom}` is not a valid path domain.

            See https://docs.latch.bio/basics/latch_urls.html for a list of valid domains.
            """).strip("\n"),
            fg="red",
        )
        raise click.exceptions.Exit(1)

    return match["account_relative"] is not None


def normalize_path(path: str, *, assume_remote: bool = False) -> str:
    path = append_scheme(path, assume_remote=assume_remote)

    if path.startswith("file://"):
        return path

    return append_domain(path)


auth = re.compile(
    r"""
    ^(
        (?P<sdk>Latch-SDK-Token) |
        (?P<execution>Latch-Execution-Token)
    )\s.*$
""",
    re.VERBOSE,
)


def get_path_error(path: str, message: str, acc_id: str) -> str:
    with_scheme = append_scheme(path)
    normalized = normalize_path(path)

    account_relative = is_account_relative(with_scheme)

    auth_header = get_auth_header()
    match = auth.match(auth_header)
    if match is None:
        auth_type = auth_header
    elif match["sdk"] is not None:
        auth_type = "SDK Token"
    else:
        auth_type = "Execution Token"

    auth_str = (
        f"{click.style(f'Authorized using:', bold=True, reset=False)} {click.style(auth_type, bold=False, reset=False)}"
        + "\n"
    )

    ws_id = user_config.workspace_id
    ws_name = user_config.workspace_name

    resolve_str = (
        f"{click.style(f'Relative path resolved to:', bold=True, reset=False)} {click.style(normalized, bold=False, reset=False)}"
        + "\n"
    )
    ws_str = (
        f"{click.style(f'Using Workspace:', bold=True, reset=False)} {click.style(ws_id, bold=False, reset=False)}"
    )
    if ws_name is not None:
        ws_str = f"{ws_str} ({ws_name})"
    ws_str += "\n"

    return click.style(
        f"""
{click.style(f'{path}: ', bold=True, reset=False)}{click.style(message, bold=False, reset=False)}
{resolve_str if account_relative else ""}{ws_str if account_relative else ""}
{auth_str}
{click.style("Check that:", bold=True, reset=False)}
{click.style("1. The target object exists", bold=False, reset=False)}
{click.style(f"2. Account ", bold=False, reset=False)}{click.style(acc_id, bold=True, reset=False)}{click.style(" has permission to view the target object", bold=False, reset=False)}
{"3. The correct workspace is selected" if account_relative else ""}

For privacy reasons, non-viewable objects and non-existent objects are indistinguishable""",
        fg="red",
    )


name = re.compile(r"^.*/(?P<name>[^/]+)/?$")


def get_name_from_path(path: str):
    match = name.match(path)
    if match is None:
        click.secho(f"`{path}` is not a valid path.", fg="red")
        raise click.exceptions.Exit(1)
    return match["name"]



from enum import Enum


class LatchPathError(RuntimeError):
    def __init__(self, message: str, remote_path: str = None, acc_id: str = None):
        super().__init__(message)
        self.message = message
        self.remote_path = remote_path
        self.acc_id = acc_id

    def __str__(self) -> str:
        return f"{self.remote_path}: {self.message}"


class LDataNodeType(str, Enum):
    account_root = "account_root"
    dir = "dir"
    obj = "obj"
    mount = "mount"
    link = "link"



import json
from datetime import date, datetime
from enum import Enum
from typing import Dict, List, Optional, Type, TypeVar, Union, cast

import gql
from dateutil.parser import parse
from latch_sdk_config.user import user_config
from latch_sdk_gql.execute import execute

from latch.registry.record import Record
from latch.registry.types import InvalidValue, RegistryPythonValue
from latch.registry.upstream_types.types import (
    ArrayType,
    PrimitiveType,
    PrimitiveTypeEnum,
    RegistryType,
)
from latch.registry.upstream_types.values import DBValue
from latch.types.directory import LatchDir
from latch.types.file import LatchFile

# todo(maximsmol): hopefully, PyLance eventually narrows `TypedDict`` unions using `in`
# then we can get rid of the casts

T = TypeVar("T")


class RegistryTransformerException(ValueError): ...


def to_python_type(registry_type: RegistryType) -> Type[RegistryPythonValue]:
    if "primitive" in registry_type:
        primitive = cast(PrimitiveType, registry_type)["primitive"]
        if primitive == "string":
            return str
        if primitive == "datetime":
            return datetime
        if primitive == "date":
            return date
        if primitive == "integer":
            return int
        if primitive == "number":
            return float
        if primitive == "blob":
            return get_blob_nodetype(registry_type)
        if primitive == "link":
            return Record
        if primitive == "enum":
            members = cast(PrimitiveTypeEnum, registry_type)["members"]
            return Enum("Enum", members)
        if primitive == "null":
            return type(None)
        if primitive == "boolean":
            return bool

        raise RegistryTransformerException(f"invalid primitive type: {primitive}")

    if "array" in registry_type:
        array = cast(ArrayType, registry_type)["array"]
        return List[to_python_type(array)]

    if "union" in registry_type:
        variants: List[Type[RegistryPythonValue]] = []
        for variant in registry_type["union"].values():
            variants.append(
                # todo(maximsmol): allow specifying the exact variant we want
                # or preserving it when round-tripping?
                to_python_type(
                    variant,
                ),
            )
        return Union[tuple(variants)]

    raise RegistryTransformerException(
        f"unknown registry type cannot be converted to a python type: {registry_type}"
    )


def to_python_literal(
    registry_literal: DBValue,
    registry_type: RegistryType,
):
    if "array" in registry_type:
        if type(registry_literal) is not list:
            raise RegistryTransformerException(
                f"{registry_literal} is not a list so it cannot be converted into a"
                " list"
            )

        return [
            to_python_literal(sub_val, registry_type["array"])
            for sub_val in registry_literal
        ]
    if "union" in registry_type:
        tag = registry_literal["tag"]
        sub_type = registry_type["union"].get(tag)
        if sub_type is None:
            raise RegistryTransformerException(
                f"{registry_literal} cannot be converted to {registry_type} because its"
                f" tag `{tag}` is not present."
            )

        return to_python_literal(registry_literal["value"], sub_type)

    if not registry_literal["valid"]:
        return InvalidValue(registry_literal["rawValue"])

    value = registry_literal["value"]

    primitive = registry_type.get("primitive")
    if primitive is None:
        raise RegistryTransformerException(
            f"cannot convert to python - malformed registry type: {registry_type}"
        )

    if primitive == "enum":
        if "members" not in registry_type:
            raise RegistryTransformerException(
                "cannot convert to python - malformed registry enum type without"
                f" members: {registry_type}"
            )

        members = registry_type["members"]
        if value not in members:
            raise RegistryTransformerException(
                f"unable to convert {value} to any of the enum members {members}"
            )

        return to_python_type(registry_type)[value]

    if primitive == "blob":
        typ = get_blob_nodetype(registry_type)

        if type(value) is not dict or "ldataNodeId" not in value:
            raise RegistryTransformerException(
                f"cannot convert non-blob value {value} to blob type"
            )

        return typ(f"latch://{value['ldataNodeId']}.node")

    if primitive == "link":
        if "sampleId" not in value:
            raise RegistryTransformerException(
                f"unable to convert {value} to Link as row id was not found"
            )

        from latch.registry.record import Record

        return Record(value["sampleId"])

    if primitive == "string":
        if isinstance(value, str):
            return value
        raise RegistryTransformerException(
            f"Cannot convert {value} to string as it is not a primitive string literal"
        )

    if primitive == "number":
        if isinstance(value, float) or isinstance(value, int):
            return float(value)

        raise RegistryTransformerException(
            f"Cannot convert {value} ({type(value)}) to float as it is not a primitive"
            " number literal"
        )

    if primitive == "integer":
        if isinstance(value, int):
            return value
        raise RegistryTransformerException(
            f"Cannot convert {value} to integer as it is not a primitive integer"
            " literal"
        )

    if primitive == "boolean":
        if isinstance(value, bool):
            return value
        raise RegistryTransformerException(
            f"Cannot convert {value} to boolean as it is not a primitive boolean"
            " literal"
        )

    if primitive == "date":
        if isinstance(value, str):
            return parse(value).date()
        raise RegistryTransformerException(
            f"Cannot convert {value} to date as it is not a primitive date literal"
        )

    if primitive == "datetime":
        if isinstance(value, str):
            return parse(value)
        raise RegistryTransformerException(
            f"Cannot convert {value} to datetime as it is not a primitive datetime"
            " literal"
        )

    if primitive == "null":
        if value is None:
            return value
        raise RegistryTransformerException(
            f"Cannot convert {value} to None as it is not a primitive null literal"
        )

    raise ValueError(
        f"primitive literal {registry_literal['value']} cannot be converted to"
        f" {registry_type}"
    )


def to_registry_literal(
    python_literal,
    registry_type: RegistryType,
) -> DBValue:
    if isinstance(python_literal, InvalidValue):
        return {"valid": False, "rawValue": python_literal.raw_value}

    if "union" in registry_type:
        errors: Dict[str, str] = {}

        for tag, sub_type in registry_type["union"].items():
            try:
                return {
                    "tag": tag,
                    "value": to_registry_literal(python_literal, sub_type),
                }
            except RegistryTransformerException as e:
                errors[str(sub_type)] = str(e)
                pass

        raise RegistryTransformerException(
            f"{python_literal} cannot be converted into any union members of"
            f" {json.dumps(registry_type, indent=2)}:\n{json.dumps(errors, indent=2)}"
        )

    if "array" in registry_type:
        sub_type = registry_type["array"]

        if type(python_literal) is not list:
            raise RegistryTransformerException(
                "unable to convert non-list python literal to registry array literal"
            )

        return [to_registry_literal(literal, sub_type) for literal in python_literal]

    if "primitive" not in registry_type:
        raise RegistryTransformerException(f"malformed registry type: {registry_type}")

    primitive = registry_type["primitive"]

    value: Optional[object] = None
    if primitive == "string":
        if not isinstance(python_literal, str):
            raise RegistryTransformerException(
                f"cannot convert non-string python literal to {primitive} registry"
                " literal"
            )

        value = python_literal
    elif primitive in ["date", "datetime"]:
        # datetime is a subclass of date
        if not isinstance(python_literal, date):
            raise RegistryTransformerException(
                f"cannot convert non-date(time) python literal to {primitive} registry"
                " literal"
            )

        value = python_literal.isoformat()
    elif primitive == "integer":
        if not isinstance(python_literal, int):
            raise RegistryTransformerException(
                "unable to convert non-int python literal to registry int literal"
            )

        value = int(python_literal)
    elif primitive == "number":
        if not (isinstance(python_literal, int) or isinstance(python_literal, float)):
            raise RegistryTransformerException(
                "unable to convert non-numeric python literal to registry number"
                " literal"
            )

        value = float(python_literal)
    elif primitive == "boolean":
        if not isinstance(python_literal, bool):
            raise RegistryTransformerException(
                "unable to convert non-boolean python literal to registry boolean"
                " literal"
            )

        value = python_literal
    elif primitive == "null":
        if not python_literal is None:
            raise RegistryTransformerException(
                "unable to convert non-None python literal to registry null literal"
            )

    elif primitive == "enum":
        members = registry_type["members"]

        if isinstance(python_literal, Enum):
            python_literal = python_literal.name
        elif not isinstance(python_literal, str):
            python_literal = str(python_literal)
        if not python_literal in members:
            raise RegistryTransformerException(
                f"unable to convert {python_literal} to registry enum with members"
                f" {', '.join(members)}"
            )

        value = python_literal
    elif primitive == "link":
        if not isinstance(python_literal, Record):
            raise RegistryTransformerException(
                "cannot convert non-record python literal to registry link"
            )

        value = {"sampleId": python_literal.id}
    elif primitive == "blob":
        if not (
            isinstance(python_literal, LatchFile)
            or isinstance(python_literal, LatchDir)
        ):
            raise RegistryTransformerException(
                "cannot convert non-blob python literal to registry blob"
            )

        ws_id = user_config.workspace_id
        if ws_id == "":
            ws_id = None

        if ws_id is None:
            data = execute(
                gql.gql("""
                query nodeIdQ($argPath: String!) {
                    ldataResolvePath(
                        path: $argPath
                    ) {
                        nodeId
                        path
                    }
                }
                """),
                {"argPath": python_literal.remote_path},
            )["ldataResolvePath"]
        else:
            data = execute(
                gql.gql("""
                query nodeIdQ($argPath: String!, $wsId: BigInt!) {
                    ldataResolvePathExt(
                        path: $argPath,
                        accId: $wsId
                    ) {
                        nodeId
                        path
                    }
                }
                """),
                {"argPath": python_literal.remote_path, "wsId": ws_id},
            )["ldataResolvePathExt"]

        if data["path"] is not None and data["path"] != "":
            # todo(maximsmol): store an invalid value instead?
            raise RegistryTransformerException(
                f"could not resolve path: {python_literal.remote_path}"
            )

        value = {"ldataNodeId": data["nodeId"]}
    else:
        raise RegistryTransformerException(f"malformed registry type: {registry_type}")

    return {"value": value, "valid": True}


def get_blob_nodetype(
    registry_type: RegistryType,
) -> Union[Type[LatchFile], Type[LatchDir]]:
    if "primitive" not in registry_type or registry_type["primitive"] != "blob":
        raise RegistryTransformerException(
            f"cannot extract blob nodetype from non-blob type"
        )

    if (
        "metadata" in registry_type
        and "nodeType" in registry_type["metadata"]
        and registry_type["metadata"]["nodeType"] == "dir"
    ):
        return LatchDir

    return LatchFile



import json
import os
import shlex
import sys
from itertools import chain
from pathlib import Path
from textwrap import dedent
from typing import Dict, Optional, Set
from urllib.parse import urlparse, urlunparse

import snakemake
import snakemake.workflow
from snakemake.deployment import singularity
from snakemake.parser import (
    INDENT,
    Benchmark,
    Include,
    Input,
    Log,
    Output,
    Params,
    Python,
    Rule,
    Ruleorder,
    Shell,
)
from snakemake.rules import Rule as RRule
from snakemake.workflow import Workflow as WWorkflow

sys.stdout.reconfigure(line_buffering=True)
sys.stderr.reconfigure(line_buffering=True)


def eprint(x: str) -> None:
    print(x, file=sys.stderr)


print_compilation = os.environ.get("LATCH_PRINT_COMPILATION", False)
if print_compilation == "1":
    print_compilation = True

data = json.loads(os.environ["LATCH_SNAKEMAKE_DATA"])
rules = data["rules"]
outputs = data["outputs"]

overwrite_config = data.get("overwrite_config", {})

old_workflow_init = WWorkflow.__init__


def new_init(self: WWorkflow, *args, **kwargs):
    kwargs["overwrite_config"] = overwrite_config
    old_workflow_init(self, *args, **kwargs)


WWorkflow.__init__ = new_init


def eprint_named_list(xs):
    eprint("      Positional:")
    for x in xs["positional"]:
        eprint(f"      {repr(x)}")

    eprint("      Keyword:")
    for k, v in xs["keyword"].items():
        eprint(f"      {k}={repr(v)}")


eprint("\n>>><<<\n")
eprint("Using LATCH_SNAKEMAKE_DATA:")
for rule in rules:
    rule_data = rules[rule]
    eprint(f"  {rule}:")

    eprint("    Inputs:")
    eprint_named_list(rule_data["inputs"])

    eprint("    Outputs:")
    eprint_named_list(rule_data["outputs"])

    eprint("    Params:")
    eprint_named_list(rule_data["params"])

    eprint("    Benchmark:")
    eprint(f"      {rule_data['benchmark']}")

    eprint("    Log:")
    eprint(f"      {rule_data['log']}")

    eprint("    Shellcmd:")
    eprint(f"      {rule_data['shellcmd']}")

eprint("\nExpected outputs:")
for x in outputs:
    eprint(repr(x))
eprint("\n>>><<<\n")

# Add a custom entrypoint rule
_real_rule_start = Rule.start


def rule_start(self, aux=""):
    prefix = ""
    if self.rulename in rules:
        outputs_str = ",\n".join(f"    {repr(x)}" for x in outputs)
        prefix = dedent(f"""
            @workflow.rule(name='latch_entrypoint', lineno=1, snakefile='workflow/Snakefile')
            @workflow.input(
            __outputs__
            )
            @workflow.norun()
            @workflow.run
            def __rule_latch_entrypoint(input, output, params, wildcards, threads, resources, log, version, rule, conda_env, container_img, singularity_args, use_singularity, env_modules, bench_record, jobid, is_shell, bench_iteration, cleanup_scripts, shadow_dir, edit_notebook, conda_base_path, basedir, runtime_sourcecache_path, __is_snakemake_rule_func=True):
                pass

            """).replace("__outputs__", outputs_str)

    yield prefix + next(_real_rule_start(self, aux))


Rule.start = rule_start


def render_annotated_str(x) -> str:
    if not isinstance(x, dict):
        return repr(x)

    value = x["value"]
    flags = dict(x["flags"])

    res = repr(value)

    if len(flags) > 1:
        raise RuntimeError(f"can only have one flag for {res} but found: {repr(flags)}")

    if "directory" in flags:
        res = f"directory({res})"

    elif "report" in flags:
        report_vals = flags.get("report", False)
        res = (
            f"report({res}, caption={repr(report_vals['caption'])},"
            f" category={report_vals['category']})"
        )

    elif "temp" in flags:
        # A temporary modifier is no different from a normal file as all files
        # are deleted on Latch after a job completes.
        # https://snakemake.readthedocs.io/en/stable/snakefiles/rules.html#protected-and-temporary-files
        del flags["temp"]

    return res


def render_annotated_str_list(xs) -> str:
    if not isinstance(xs, list):
        return render_annotated_str(xs)

    els = ", ".join(render_annotated_str(x) for x in xs)
    return f"[{els}]"


def emit_overrides(self, token):
    cur_data = rules[self.rulename]
    eprint(f"\nOverriding {self.rulename} {self.__class__.__name__}")

    if isinstance(self, Input):
        xs = cur_data["inputs"]
    elif isinstance(self, Output):
        xs = cur_data["outputs"]
    elif isinstance(self, Params):
        xs = cur_data["params"]
    elif isinstance(self, Benchmark):
        xs = {"positional": [cur_data["benchmark"]], "keyword": {}}
    elif isinstance(self, Log):
        xs = {"positional": [cur_data["log"]], "keyword": {}}
    else:
        raise ValueError(f"tried to emit overrides for unknown state: {type(self)}")

    if (
        isinstance(self, Output)
        and len(xs["positional"]) > 0
        and xs["positional"][0].get("flags") is not None
        and "multiext" in xs["positional"][0].get("flags")
    ):
        filename = repr(xs["positional"][0]["flags"]["multiext"])
        exts = [repr("." + x["value"].split(".")[-1]) for x in xs["positional"]]
        positional_data = (f"multiext({filename},{','.join(exts)})",)
    else:
        positional_data = (render_annotated_str_list(x) for x in xs["positional"])

    modifier_fn = render_annotated_str_list
    if isinstance(self, Params):
        modifier_fn = repr

    keyword_data = (f"{k}={modifier_fn(v)}" for k, v in xs["keyword"].items())
    data = chain(positional_data, keyword_data)

    for x in data:
        eprint(f"  {x}")

        yield x, token
        yield ",", token
        yield "\n", token

        # we'll need to re-indent the commented-out originals too
        yield INDENT * self.base_indent, token


emitted_overrides_per_type: Dict[str, Set[str]] = {}


def skipping_block_content(self, token):
    if self.rulename not in rules:
        return

    emitted_overrides = emitted_overrides_per_type.setdefault(
        type(self).__name__, set()
    )
    if self.rulename in emitted_overrides:
        return

    yield from emit_overrides(self, token)
    emitted_overrides.add(self.rulename)


def block_content_with_print_compilation(self, token):
    if print_compilation:
        yield f"{token.string}, print_compilation=True", token
    else:
        yield token.string, token


def empty_generator(self, token):
    return
    yield


Input.block_content = skipping_block_content
Output.block_content = skipping_block_content
Params.block_content = skipping_block_content
Benchmark.block_content = skipping_block_content
Log.block_content = skipping_block_content
Ruleorder.block_content = empty_generator
Include.block_content = block_content_with_print_compilation


class SkippingRule(Rule):
    def start(self, aux=""):
        if self.rulename not in rules:
            # Rules can be nested in conditional statements:
            #
            # if (<condition>):
            #   rule A:
            #       <stuff>
            #
            # We want correct python code if we remove them.
            yield "..."
            return

        yield from super().start(aux)

    def end(self):
        if self.rulename not in rules:
            return

        yield from super().end()

    def block_content(self, token):
        if self.rulename not in rules:
            return

        yield from super().block_content(token)


class SkippingCheckpoint(SkippingRule):
    def start(self):
        yield from super().start(aux=", checkpoint=True")


Python.subautomata["rule"] = SkippingRule
Python.subautomata["checkpoint"] = SkippingCheckpoint


class ReplacingShell(Shell):
    def __init__(self, snakefile, rulename, base_indent=0, dedent=0, root=True):
        if rulename in rules:
            cmd: str = rules[rulename]["shellcmd"]
            self.overwrite_cmd = cmd.replace("{", "{{").replace("}", "}}")

        super().__init__(snakefile, rulename, base_indent, dedent, root)


SkippingRule.subautomata["shell"] = ReplacingShell


def get_wildcards(self, requested_output, wildcards_dict=None):
    return wildcards_dict


RRule.get_wildcards = get_wildcards


def get_docker_cmd(
    img_path: str,
    cmd: str,
    args: str = "",
    quiet: bool = False,
    envvars: Optional[Dict[str, str]] = None,
    shell_executable: Optional[str] = None,
    container_workdir: Optional[str] = None,
    is_python_script: bool = False,
) -> str:
    if shell_executable is None:
        shell_executable = "sh"
    else:
        # Ensure to just use the name of the executable, not a path,
        # because we cannot be sure where it is located in the container.
        shell_executable = Path(shell_executable).name

    workdir = container_workdir if container_workdir is not None else "/latch"
    if '"' in workdir:
        raise ValueError("container_workdir cannot contain double quotes")

    docker_cmd = ["docker", "run"]
    if envvars is not None:
        for k, v in envvars.items():
            docker_cmd.extend(["--env", f"{k}={v}"])

    # support for containers currently only works with relative input and output
    # paths. This preserves the behavior before the monkey path
    docker_cmd.extend([
        "--workdir",
        workdir,
        "--mount",
        f'type=bind,src=.,"target={workdir}"',
    ])

    if args != "":
        docker_cmd.extend(shlex.split(args))

    docker_cmd.extend([
        img_path,
        shell_executable,
        "-c",
        cmd,
    ])

    return shlex.join(docker_cmd)


@property
def docker_path(self):
    parsed = urlparse(self.url)
    if parsed.scheme != "" and parsed.scheme != "docker":
        raise ValueError("Only docker images are supported")
    path = urlunparse(parsed._replace(scheme=""))
    return path[2:] if path.startswith("//") else path


# Monkey-patch snakemake to use docker container runtime instead of singularity
# because singularity does not work inside sysbox.
singularity.Image.pull = lambda self, dry_run: None
singularity.Image.path = docker_path
singularity.shellcmd = get_docker_cmd


# Run snakemake
snakemake.main()



import hashlib
import importlib
import json
import sys
import typing
from dataclasses import dataclass, fields, is_dataclass
from pathlib import Path
from textwrap import dedent
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from urllib.parse import urlparse

import click
import snakemake
import snakemake.io
import snakemake.jobs
from flytekit.configuration import SerializationSettings
from flytekit.core import constants as _common_constants
from flytekit.core.base_task import TaskMetadata
from flytekit.core.class_based_resolver import ClassStorageTaskResolver
from flytekit.core.context_manager import FlyteContextManager
from flytekit.core.docstring import Docstring
from flytekit.core.interface import Interface, transform_interface_to_typed_interface
from flytekit.core.node import Node
from flytekit.core.promise import NodeOutput, Promise
from flytekit.core.python_auto_container import (
    DefaultTaskResolver,
    PythonAutoContainerTask,
)
from flytekit.core.type_engine import TypeEngine, TypeTransformerFailedError
from flytekit.core.workflow import (
    WorkflowBase,
    WorkflowFailurePolicy,
    WorkflowMetadata,
    WorkflowMetadataDefaults,
)
from flytekit.exceptions import scopes as exception_scopes
from flytekit.models import interface as interface_models
from flytekit.models import literals as literals_models
from flytekit.models import task as _task_models
from flytekit.models import types as type_models
from flytekit.models.core.types import BlobType
from flytekit.models.literals import Blob, BlobMetadata, Literal, LiteralMap, Scalar
from flytekitplugins.pod.task import (
    _PRIMARY_CONTAINER_NAME_FIELD,
    Pod,
    PodFunctionTask,
    _sanitize_resource_name,
)
from kubernetes.client import ApiClient
from kubernetes.client.models import (
    V1Container,
    V1EnvVar,
    V1PodSpec,
    V1ResourceRequirements,
    V1Toleration,
)
from snakemake.dag import DAG
from snakemake.jobs import GroupJob, Job
from typing_extensions import Annotated, TypeAlias, TypedDict

import latch.types.metadata as metadata
from latch.resources.tasks import _get_large_gpu_pod, _get_small_gpu_pod, custom_task
from latch.types.directory import LatchDir
from latch.types.file import LatchFile
from latch.types.metadata import FileMetadata, SnakemakeFileParameter
from latch_cli.snakemake.config.utils import is_list_type, is_primitive_type, type_repr
from latch_cli.snakemake.utils import reindent

from ..utils import identifier_from_str, identifier_suffix_from_str

SnakemakeInputVal: TypeAlias = snakemake.io._IOFile


# old snakemake did not have encode_target_jobs_cli_args
def jobs_cli_args(
    jobs: Iterable[Job],
) -> Generator[str, None, None]:
    for x in jobs:
        wildcards = ",".join(
            f"{key}={value}" for key, value in x.wildcards_dict.items()
        )
        yield f"{x.rule.name}:{wildcards}"


T = TypeVar("T")


@dataclass
class JobOutputInfo:
    jobid: str
    output_param_name: str
    type_: Union[Type[LatchFile], Type[LatchDir]]


def task_fn_placeholder(): ...


def variable_name_for_file(file: snakemake.io.AnnotatedString):
    if file[0] == "/":
        return f"a_{identifier_suffix_from_str(file)}"

    return f"r_{identifier_suffix_from_str(file)}"


@dataclass
class RemoteFile:
    local_path: str
    remote_path: str


def snakemake_dag_to_interface(
    dag: DAG,
    wf_name: str,
    docstring: Optional[Docstring] = None,
    local_to_remote_path_mapping: Optional[Dict[str, str]] = None,
) -> Tuple[Interface, LiteralMap, List[RemoteFile]]:
    outputs: Dict[str, Union[Type[LatchFile], Type[LatchDir]]] = {}
    for target in dag.targetjobs:
        for desired in target.input:
            param = variable_name_for_file(desired)

            jobs: List[snakemake.jobs.Job] = dag.file2jobs(desired)
            producer_out: snakemake.io._IOFile = next(x for x in jobs[0].output)
            if producer_out.is_directory:
                outputs[param] = LatchDir
            else:
                outputs[param] = LatchFile

    literals: Dict[str, Literal] = {}
    inputs: Dict[str, Tuple[Type[LatchFile], None]] = {}
    return_files: List[RemoteFile] = []
    for job in dag.jobs:
        dep_outputs = []
        for dep, dep_files in dag.dependencies[job].items():
            for o in dep.output:
                if o in dep_files:
                    dep_outputs.append(o)
            for o in dep.log:
                if o in dep_files:
                    dep_outputs.append(o)

        for x in job.input:
            if x not in dep_outputs:
                param = variable_name_for_file(x)
                inputs[param] = (LatchFile, None)

                remote_path = (
                    Path("/.snakemake_latch") / "workflows" / wf_name / "inputs" / x
                )
                use_original_remote_path: bool = (
                    local_to_remote_path_mapping is not None
                    and x in local_to_remote_path_mapping
                )

                if use_original_remote_path:
                    remote_path = local_to_remote_path_mapping.get(x)

                remote_url = (
                    urlparse(str(remote_path))._replace(scheme="latch").geturl()
                )

                if not use_original_remote_path:
                    return_files.append(
                        RemoteFile(local_path=x, remote_path=remote_url)
                    )

                # it is possible for two different input files to have the same parameter name
                # as long as they are defined in different rules. This can lead to name collisions
                # because we use a global mapping (across all rules) of param_name -> input_file.
                if param in literals and literals[param].scalar.blob.uri != remote_url:
                    click.secho(
                        dedent(f"""
                        Name collision detected for {job.name}. The following two input
                        parameters have the same name but different remote paths:
                        {param} -> {remote_url}
                        {param} -> {literals[param].scalar.blob.uri}
                        """),
                        fg="red",
                    )
                    sys.exit(1)

                literals[param] = Literal(
                    scalar=Scalar(
                        blob=Blob(
                            metadata=BlobMetadata(
                                type=BlobType(
                                    format="",
                                    dimensionality=BlobType.BlobDimensionality.SINGLE,
                                )
                            ),
                            uri=remote_url,
                        ),
                    )
                )

    meta = metadata.LatchMetadata(
        display_name=wf_name,
        author=metadata.LatchAuthor(name="Latch Snakemake JIT"),
        parameters={k: metadata.LatchParameter(display_name=k) for k in inputs.keys()},
    )

    return (
        Interface(
            inputs,
            outputs,
            docstring=Docstring(f"{wf_name}\n\nSample Description\n\n" + str(meta)),
        ),
        LiteralMap(literals=literals),
        return_files,
    )


def binding_data_from_python(
    expected_literal_type: type_models.LiteralType,
    t_value: typing.Any,
    t_value_type: Optional[Type] = None,
) -> Optional[literals_models.BindingData]:
    if isinstance(t_value, Promise):
        if not t_value.is_ready:
            return literals_models.BindingData(promise=t_value.ref)


def binding_from_python(
    var_name: str,
    expected_literal_type: type_models.LiteralType,
    t_value: typing.Any,
    t_value_type: Type,
) -> literals_models.Binding:
    binding_data = binding_data_from_python(
        expected_literal_type, t_value, t_value_type
    )
    return literals_models.Binding(var=var_name, binding=binding_data)


def transform_type(
    x: Type, description: Optional[str] = None
) -> interface_models.Variable:
    return interface_models.Variable(
        type=TypeEngine.to_literal_type(x), description=description
    )


def transform_types_in_variable_map(
    variable_map: Dict[str, Type],
    descriptions: Dict[str, str] = {},
) -> Dict[str, interface_models.Variable]:
    res = {}
    if variable_map:
        for k, v in variable_map.items():
            res[k] = transform_type(v, descriptions.get(k, k))
    return res


def interface_to_parameters(
    interface: Optional[Interface],
) -> interface_models.ParameterMap:
    if interface is None or interface.inputs_with_defaults is None:
        return interface_models.ParameterMap({})

    if interface.docstring is None:
        inputs_vars = transform_types_in_variable_map(interface.inputs)
    else:
        inputs_vars = transform_types_in_variable_map(
            interface.inputs, interface.docstring.input_descriptions
        )

    params: Dict[str, interface_models.Parameter] = {}
    for k, v in inputs_vars.items():
        val, default = interface.inputs_with_defaults[k]
        required = default is None
        default_lv = None

        ctx = FlyteContextManager.current_context()
        if default is not None:
            default_lv = TypeEngine.to_literal(
                ctx, default, python_type=interface.inputs[k], expected=v.type
            )

        params[k] = interface_models.Parameter(
            var=v, default=default_lv, required=required
        )
    return interface_models.ParameterMap(params)


class JITRegisterWorkflow(WorkflowBase, ClassStorageTaskResolver):
    out_parameter_name = "o0"  # must be "o0"

    def __init__(self, cache_tasks: bool = False):
        self.cache_tasks = cache_tasks

        assert metadata._snakemake_metadata is not None

        parameter_metadata = metadata._snakemake_metadata.parameters
        display_name = metadata._snakemake_metadata.display_name
        name = metadata._snakemake_metadata.name
        self.file_metadata = metadata._snakemake_metadata.file_metadata

        docstring = Docstring(
            f"{display_name}\n\nSample Description\n\n"
            + str(metadata._snakemake_metadata)
        )
        python_interface = Interface(
            {k: (v.type, v.default) for k, v in parameter_metadata.items()},
            {self.out_parameter_name: bool},
            docstring=docstring,
        )
        self.parameter_metadata = parameter_metadata
        if metadata._snakemake_metadata.output_dir is not None:
            self.remote_output_url = metadata._snakemake_metadata.output_dir.remote_path
        else:
            self.remote_output_url = None

        workflow_metadata = WorkflowMetadata(
            on_failure=WorkflowFailurePolicy.FAIL_IMMEDIATELY
        )
        name = f"{name}_jit_register"
        workflow_metadata_defaults = WorkflowMetadataDefaults(False)
        super().__init__(
            name=name,
            workflow_metadata=workflow_metadata,
            workflow_metadata_defaults=workflow_metadata_defaults,
            python_interface=python_interface,
        )

    def get_fn_interface(
        self, decorator_name="small_task", fn_name: Optional[str] = None
    ):
        if fn_name is None:
            fn_name = self.name

        params: List[str] = []
        for param, t in self.python_interface.inputs.items():
            params.append(
                reindent(
                    rf"""
                    {param}: {type_repr(t, add_namespace=True)}
                    """,
                    1,
                ).rstrip()
            )

        params_str = ",\n".join(params)

        return reindent(
            rf"""
            @{decorator_name}
            def {fn_name}(
            __params__
            ) -> bool:
            """,
            0,
        ).replace("__params__", params_str)

    def get_fn_return_stmt(self):
        return reindent(
            rf"""
            return True
            """,
            1,
        )

    def _param_path_str(self, param_path: List[Union[str, int]]) -> str:
        return "][".join(map(repr, param_path))

    def get_file_download_code(
        self,
        t: Type,
        param: str,
        path: Path,
        download: bool,
        config: bool,
        param_path: str,
    ) -> str:
        code_block = ""

        param_name = identifier_from_str(param)
        touch_str = f"{param}._create_imposters()"
        if download:
            touch_str = (
                f'print(f"Downloading {param}: {{{param}.remote_path}}");'
                f" Path({param}).resolve()"
            )

        code_block += dedent(rf"""
            {param_name}_dst_p = Path("{path}")
            {touch_str}
            {param_name}_p = Path({param}.path)
            print(f"  {{file_name_and_size({param_name}_p)}}")
            """)

        if t is LatchDir:
            code_block += f"{param_name}_p.iterdir()\n"

        code_block += dedent(rf"""
            print(f"Moving {param} to {{{param_name}_dst_p}}")
            update_mapping({param_name}_p, {param_name}_dst_p, {param}.remote_path, local_to_remote_path_mapping)
            check_exists_and_rename(
                {param_name}_p,
                {param_name}_dst_p
            )
            """)

        if config:
            code_block += dedent(rf"""
                print(f"Saving parameter value {param} = {str(path)}")
                overwrite_config[{self._param_path_str(param_path)}] = {repr(str(path))}
                """)

        return code_block

    def get_param_code(
        self,
        t: Type,
        param: str,
        file_meta: FileMetadata,
        param_path: List[Union[str, int]],
    ) -> str:
        if get_origin(t) is Union:
            args = get_args(t)
            assert len(args) > 0
            code_block = f"if {param} is not None:\n"
            return code_block + reindent(
                self.get_param_code(args[0], param, file_meta, param_path), 1
            )

        param_meta = self.parameter_metadata.get(param)
        # support workflows that use the legacy SnakemakeFileParameter
        if isinstance(param_meta, SnakemakeFileParameter):
            assert t in (
                LatchFile,
                LatchDir,
            ), "SnakemakeFileParameter must have type LatchFile or LatchDir"

            code_block = ""
            code_block += self.get_file_download_code(
                t,
                param,
                param_meta.path,
                param_meta.download,
                param_meta.config,
                param_path,
            )

            return code_block

        if file_meta is None or file_meta.get(param) is None or is_primitive_type(t):
            return dedent(f"""
                if {param} is not None:
                    print(f"Saving parameter value {param} = get_parameter_json_value({param})")
                    overwrite_config[{self._param_path_str(param_path)}] = get_parameter_json_value({param})
                """)

        if get_origin(t) is Annotated:
            args = get_args(t)
            assert len(args) > 0
            return self.get_param_code(args[0], param, file_meta, param_path)

        meta = file_meta.get(param)
        if t in (LatchFile, LatchDir):
            return self.get_file_download_code(
                t,
                param,
                meta.path,
                meta.download,
                meta.config,
                param_path,
            )

        code_blocks: List[str] = []
        if is_list_type(t):
            args = get_args(t)
            if len(args) == 0:
                click.secho(
                    dedent(f"""
                    Generic Lists are not supported - please specify a subtype, e.g."
                    List[LatchFile]",
                    """),
                    fg="red",
                )
                raise click.exceptions.Exit(1)

            sub_typ = args[0]
            code_blocks.append(
                f"overwrite_config[{self._param_path_str(param_path)}] = [None for"
                f" _ in range({len(meta)})]\n"
            )

            for i, m in enumerate(meta):
                sub_param = f"{param}[{i}]"
                code_blocks.append(
                    self.get_param_code(
                        sub_typ,
                        sub_param,
                        {sub_param: m},
                        [*param_path, i],
                    )
                )
        else:
            assert is_dataclass(t)
            code_blocks.append(
                f"overwrite_config[{self._param_path_str(param_path)}] = {{}}"
            )
            for field in fields(t):
                sub_param = f"{param}.{field.name}"
                code_blocks.append(
                    self.get_param_code(
                        field.type,
                        sub_param,
                        {sub_param: meta.get(field.name)},
                        param_path + [field.name],
                    )
                )

        return "\n".join(code_blocks)

    def get_fn_code(
        self,
        snakefile_path: str,
        image_name: str,
        remote_output_url: Optional[str],
    ):
        task_name = f"{self.name}_task"

        code_block = self.get_fn_interface(fn_name=task_name)

        code_block += reindent(
            rf"""
            overwrite_config = {{}}
            local_to_remote_path_mapping = {{}}

            """,
            1,
        )

        for param, t in self.python_interface.inputs.items():
            code_block += reindent(
                self.get_param_code(t, param, self.file_metadata, [param]) + "\n",
                1,
            )

        code_block += reindent(
            rf"""
            image_name = "{image_name}"
            snakefile = Path("{snakefile_path}")

            lp = LatchPersistence()
            pkg_root = Path.cwd()
            """,
            1,
        )

        code_block += reindent(
            r"""

            dry_run = os.environ.get("LATCH_SNAKEMAKE_DRY_RUN")
            if dry_run is not None:
                token = None
                version = None
                jit_wf_version = "0.0.0-dry"
                jit_exec_display_name = "jit-dry-run"
                account_id = None
            else:
                exec_id_hash = hashlib.sha1()
                token = os.environ["FLYTE_INTERNAL_EXECUTION_ID"]
                exec_id_hash.update(token.encode("utf-8"))
                version = exec_id_hash.hexdigest()[:16]

                jit_wf_version = os.environ["FLYTE_INTERNAL_TASK_VERSION"]
                res = execute(
                    gql.gql('''
                    query executionCreatorsByToken($token: String!) {
                        executionCreatorByToken(token: $token) {
                            flytedbId
                            info {
                                displayName
                            }
                            accountInfoByCreatedBy {
                                id
                            }
                        }
                    }
                    '''),
                    {"token": token},
                )["executionCreatorByToken"]

                jit_exec_display_name = res["info"]["displayName"]
                account_id = res["accountInfoByCreatedBy"]["id"]
            """,
            1,
        )

        code_block += reindent(
            rf"""
            print(f"JIT Workflow Version: {{jit_wf_version}}")
            print(f"JIT Execution Display Name: {{jit_exec_display_name}}")

            wf = extract_snakemake_workflow(
                pkg_root,
                snakefile,
                jit_wf_version,
                jit_exec_display_name,
                local_to_remote_path_mapping,
                overwrite_config,
                {self.cache_tasks},
            )
            wf_name = wf.name
            generate_snakemake_entrypoint(wf, pkg_root, snakefile, {repr(remote_output_url)}, overwrite_config)

            entrypoint_remote = f"latch:///.snakemake_latch/workflows/{{wf_name}}/{{jit_wf_version}}/{{jit_exec_display_name}}/entrypoint.py"
            lp.upload("latch_entrypoint.py", entrypoint_remote)
            print(f"latch_entrypoint.py -> {{entrypoint_remote}}")
            """,
            1,
        )

        code_block += reindent(
            r"""

            if dry_run is None:
                headers = {
                    "Authorization": f"Latch-Execution-Token {token}",
                }

                temp_dir = tempfile.TemporaryDirectory()
                with Path(temp_dir.name).resolve() as td:
                    serialize_snakemake(wf, td, image_name, config.dkr_repo)

                    protos = _recursive_list(td)
                    reg_resp = register_serialized_pkg(protos, None, version, account_id)
                    _print_reg_resp(reg_resp, image_name)

                wf_spec_remote = f"latch:///.snakemake_latch/workflows/{wf_name}/{version}/spec"
                spec_dir = Path("spec")
                for x_dir in spec_dir.iterdir():
                    if not x_dir.is_dir():
                        dst = f"{wf_spec_remote}/{x_dir.name}"
                        print(f"{x_dir} -> {dst}")
                        lp.upload(str(x_dir), dst)
                        print("  done")
                        continue

                    for x in x_dir.iterdir():
                        dst = f"{wf_spec_remote}/{x_dir.name}/{x.name}"
                        print(f"{x} -> {dst}")
                        lp.upload(str(x), dst)
                        print("  done")

                class _WorkflowInfoNode(TypedDict):
                    id: str


                nodes: Optional[List[_WorkflowInfoNode]] = None
                while True:
                    time.sleep(1)
                    print("Getting Workflow Data:", end=" ")
                    nodes = execute(
                        gql.gql('''
                        query workflowQuery($name: String, $ownerId: BigInt, $version: String) {
                            workflowInfos(condition: { name: $name, ownerId: $ownerId, version: $version}) {
                                nodes {
                                    id
                                }
                            }
                        }
                        '''),
                        {"name": wf_name, "version": version, "ownerId": account_id},
                    )["workflowInfos"]["nodes"]

                    if not nodes:
                        print("Failed. Trying again.")
                    else:
                        print("Succeeded.")
                        break


                if len(nodes) > 1:
                    raise ValueError(
                        "Invariant violated - more than one workflow identified for unique combination"
                        " of {wf_name}, {version}, {account_id}"
                    )

                print(nodes)

                for file in wf.return_files:
                    print(f"Uploading {file.local_path} -> {file.remote_path}")
                    lp.upload(file.local_path, file.remote_path)

                wf_id = nodes[0]["id"]
                params = gpjson.MessageToDict(wf.literal_map.to_flyte_idl()).get("literals", {})

                print(params)

                _interface_request = {
                    "workflow_id": wf_id,
                    "params": params,
                    "snakemake_jit": True,
                }

                response = requests.post(urljoin(config.nucleus_url, "/api/create-execution"), headers=headers, json=_interface_request)
                print(response.json())
            else:
                print("Dry run successful. Exiting.")
            """,
            1,
        )
        code_block += self.get_fn_return_stmt()
        return code_block


class SnakemakeWorkflow(WorkflowBase, ClassStorageTaskResolver):
    def __init__(
        self,
        dag: DAG,
        jit_wf_version: str,
        jit_exec_display_name: str,
        local_to_remote_path_mapping: Optional[Dict[str, str]] = None,
        cache_tasks: bool = False,
    ):
        assert metadata._snakemake_metadata is not None
        name = metadata._snakemake_metadata.name
        self.jit_wf_version = jit_wf_version
        self.jit_exec_display_name = jit_exec_display_name
        self.cores = metadata._snakemake_metadata.cores

        assert name is not None

        python_interface, literal_map, return_files = snakemake_dag_to_interface(
            dag,
            name,
            None,
            local_to_remote_path_mapping,
        )

        self.literal_map = literal_map
        self.return_files = return_files
        self._input_parameters = None
        self._dag = dag
        self._cache_tasks = cache_tasks
        self._docker_metadata = metadata._snakemake_metadata.docker_metadata
        self._use_conda = (
            metadata._snakemake_metadata.env_config is not None
            and metadata._snakemake_metadata.env_config.use_conda
        )
        self._use_container = (
            metadata._snakemake_metadata.env_config is not None
            and metadata._snakemake_metadata.env_config.use_container
        )
        self._container_args = (
            []
            if metadata._snakemake_metadata.env_config is None
            else metadata._snakemake_metadata.env_config.container_args
        )
        self.snakemake_tasks: List[SnakemakeJobTask] = []

        workflow_metadata = WorkflowMetadata(
            on_failure=WorkflowFailurePolicy.FAIL_IMMEDIATELY
        )
        workflow_metadata_defaults = WorkflowMetadataDefaults(False)
        super().__init__(
            name=name,
            workflow_metadata=workflow_metadata,
            workflow_metadata_defaults=workflow_metadata_defaults,
            python_interface=python_interface,
        )

    def compile(self, **kwargs):
        self._input_parameters = interface_to_parameters(self.python_interface)

        GLOBAL_START_NODE = Node(
            id=_common_constants.GLOBAL_INPUT_NODE_ID,
            metadata=None,
            bindings=[],
            upstream_nodes=[],
            flyte_entity=None,
        )

        node_map: Dict[str, Node] = {}

        target_files = [x for job in self._dag.targetjobs for x in job.input]

        node_id = 0
        for layer in self._dag.toposorted():
            for job in layer:
                assert isinstance(job, snakemake.jobs.Job)
                is_target = False

                if job in self._dag.targetjobs:
                    continue

                target_file_for_output_param: Dict[str, str] = {}
                target_file_for_input_param: Dict[str, str] = {}

                python_outputs: Dict[str, Union[Type[LatchFile], Type[LatchDir]]] = {}
                for x in job.output:
                    assert isinstance(x, SnakemakeInputVal)

                    if x in target_files:
                        is_target = True
                    param = variable_name_for_file(x)
                    target_file_for_output_param[param] = x

                    if x.is_directory:
                        python_outputs[param] = LatchDir
                    else:
                        python_outputs[param] = LatchFile

                for x in job.log:
                    assert isinstance(x, SnakemakeInputVal)

                    if x in target_files:
                        is_target = True
                    param = variable_name_for_file(x)
                    target_file_for_output_param[param] = x

                    if x.is_directory:
                        python_outputs[param] = LatchDir
                    else:
                        python_outputs[param] = LatchFile

                dep_outputs: Dict[SnakemakeInputVal, JobOutputInfo] = {}
                for dep, dep_files in self._dag.dependencies[job].items():
                    for o in dep.output:
                        if o in dep_files:
                            assert isinstance(o, SnakemakeInputVal)

                            dep_outputs[o] = JobOutputInfo(
                                jobid=dep.jobid,
                                output_param_name=variable_name_for_file(o),
                                type_=LatchDir if o.is_directory else LatchFile,
                            )

                    for o in dep.log:
                        if o in dep_files:
                            assert isinstance(o, SnakemakeInputVal)

                            dep_outputs[o] = JobOutputInfo(
                                jobid=dep.jobid,
                                output_param_name=variable_name_for_file(o),
                                type_=LatchDir if o.is_directory else LatchFile,
                            )

                python_inputs: Dict[str, Union[Type[LatchFile], Type[LatchDir]]] = {}
                promise_map: Dict[str, JobOutputInfo] = {}
                for x in job.input:
                    param = variable_name_for_file(x)
                    target_file_for_input_param[param] = x

                    dep_out = dep_outputs.get(x)

                    python_inputs[param] = LatchFile

                    if dep_out is not None:
                        python_inputs[param] = dep_out.type_
                        promise_map[param] = dep_out

                interface = Interface(python_inputs, python_outputs, docstring=None)
                task = SnakemakeJobTask(
                    wf=self,
                    job=job,
                    inputs=python_inputs,
                    outputs=python_outputs,
                    target_file_for_input_param=target_file_for_input_param,
                    target_file_for_output_param=target_file_for_output_param,
                    is_target=is_target,
                    interface=interface,
                )

                if getattr(task, "_metadata") is None:
                    task._metadata = TaskMetadata()

                if self._cache_tasks:
                    task._metadata.cache = True
                    task._metadata.cache_serialize = True

                    hash = hashlib.new("sha256")
                    hash.update(job.properties().encode())
                    if job.is_script:
                        hash.update(Path(job.rule.script).read_bytes())

                    task._metadata.cache_version = hash.hexdigest()

                self.snakemake_tasks.append(task)

                typed_interface = transform_interface_to_typed_interface(interface)
                assert typed_interface is not None

                bindings: List[literals_models.Binding] = []
                for k in interface.inputs:
                    var = typed_interface.inputs[k]
                    if var.description in promise_map:
                        job_output_info = promise_map[var.description]
                        promise_to_bind = Promise(
                            var=k,
                            val=NodeOutput(
                                node=node_map[job_output_info.jobid],
                                var=job_output_info.output_param_name,
                            ),
                        )
                    else:
                        promise_to_bind = Promise(
                            var=k,
                            val=NodeOutput(node=GLOBAL_START_NODE, var=k),
                        )
                    bindings.append(
                        binding_from_python(
                            var_name=k,
                            expected_literal_type=var.type,
                            t_value=promise_to_bind,
                            t_value_type=interface.inputs[k],
                        )
                    )

                upstream_nodes = []
                for x in self._dag.dependencies[job].keys():
                    if x.jobid in node_map:
                        upstream_nodes.append(node_map[x.jobid])

                node = Node(
                    id=f"n{node_id}",
                    metadata=task.construct_node_metadata(),
                    bindings=sorted(bindings, key=lambda b: b.var),
                    upstream_nodes=upstream_nodes,
                    flyte_entity=task,
                )
                node_map[job.jobid] = node

                node_id += 1

        bindings: Dict[str, literals_models.Binding] = {}
        for target in self._dag.targetjobs:
            for out_file in target.input:
                out_param_name = variable_name_for_file(out_file)
                if out_param_name in bindings:
                    # rahul: Flyte will error if there are mulitple
                    # bindings for the same output
                    continue

                upstream_id, upstream_var = self.find_upstream_node_matching_file(
                    target, out_file
                )
                promise_to_bind = Promise(
                    var=out_param_name,
                    val=NodeOutput(node=node_map[upstream_id], var=upstream_var),
                )
                t = self.python_interface.outputs[out_param_name]
                b = binding_from_python(
                    out_param_name,
                    self.interface.outputs[out_param_name].type,
                    promise_to_bind,
                    t,
                )
                bindings[out_param_name] = b

        self._nodes = list(node_map.values())
        self._output_bindings = list(bindings.values())

    def find_upstream_node_matching_file(self, job: snakemake.jobs.Job, out_file: str):
        for depen, files in self._dag.dependencies[job].items():
            for f in files:
                if f == out_file:
                    return depen.jobid, variable_name_for_file(f)

        raise RuntimeError(f"could not find upstream node for output file: {out_file}")

    def execute(self, **kwargs):
        return exception_scopes.user_entry_point(self._workflow_function)(**kwargs)


def build_jit_register_wrapper(cache_tasks: bool = False) -> JITRegisterWorkflow:
    wrapper_wf = JITRegisterWorkflow(cache_tasks)
    out_parameter_name = wrapper_wf.out_parameter_name

    python_interface = wrapper_wf.python_interface
    wrapper_wf._input_parameters = interface_to_parameters(python_interface)

    GLOBAL_START_NODE = Node(
        id=_common_constants.GLOBAL_INPUT_NODE_ID,
        metadata=None,
        bindings=[],
        upstream_nodes=[],
        flyte_entity=None,
    )
    task_interface = Interface(
        python_interface.inputs, python_interface.outputs, docstring=None
    )

    primary_container = V1Container(name="primary")
    resources = V1ResourceRequirements(
        requests={
            "cpu": "100m",
            "memory": "200Mi",
            "ephemeral-storage": "50Gi",
        },
        limits={
            "cpu": "200m",
            "memory": "400Mi",
            "ephemeral-storage": "100Gi",
        },
    )
    primary_container.resources = resources

    task_config = Pod(
        pod_spec=V1PodSpec(
            node_selector={
                "node_group_name": "prion-jit-tasks",
            },
            containers=[primary_container],
            tolerations=[
                V1Toleration(effect="NoSchedule", key="ng", value="jit-tasks")
            ],
        ),
        primary_container_name="primary",
    )

    task = PodFunctionTask(
        task_config=task_config,
        task_function=task_fn_placeholder,
        task_resolver=JITRegisterWorkflowResolver(),
    )
    # use custom task name and interface instead of the one generated from the fn_placeholder
    task._name = f"{wrapper_wf.name}_task"
    task._python_interface = task_interface

    typed_interface = transform_interface_to_typed_interface(python_interface)
    assert typed_interface is not None
    task._interface = typed_interface

    task_bindings: List[literals_models.Binding] = []
    for k in python_interface.inputs:
        var = typed_interface.inputs[k]
        promise_to_bind = Promise(
            var=k,
            val=NodeOutput(node=GLOBAL_START_NODE, var=k),
        )
        task_bindings.append(
            binding_from_python(
                var_name=k,
                expected_literal_type=var.type,
                t_value=promise_to_bind,
                t_value_type=python_interface.inputs[k],
            )
        )
    task_node = Node(
        id="n0",
        metadata=task.construct_node_metadata(),
        bindings=sorted(task_bindings, key=lambda b: b.var),
        upstream_nodes=[],
        flyte_entity=task,
    )

    promise_to_bind = Promise(
        var=out_parameter_name,
        val=NodeOutput(node=task_node, var=out_parameter_name),
    )
    t = python_interface.outputs[out_parameter_name]
    output_binding = binding_from_python(
        out_parameter_name,
        bool,
        promise_to_bind,
        t,
    )

    wrapper_wf._nodes = [task_node]
    wrapper_wf._output_bindings = [output_binding]
    return wrapper_wf


class AnnotatedStrJson(TypedDict):
    value: str
    flags: Dict[str, bool]


MaybeAnnotatedStrJson: TypeAlias = Union[str, AnnotatedStrJson]


def annotated_str_to_json(
    x: Union[str, snakemake.io._IOFile, snakemake.io.AnnotatedString]
) -> MaybeAnnotatedStrJson:
    if not isinstance(x, (snakemake.io.AnnotatedString, snakemake.io._IOFile)):
        return x

    flags = dict(x.flags.items())
    if "report" in flags:
        report = flags["report"]
        flags["report"] = {
            "caption": report.caption.get_filename(),
            "category": report.category,
        }

    return {"value": str(x), "flags": flags}


IONamedListItem = Union[MaybeAnnotatedStrJson, List[MaybeAnnotatedStrJson]]


class NamedListJson(TypedDict):
    positional: List[IONamedListItem]
    keyword: Dict[str, IONamedListItem]


def named_list_to_json(xs: snakemake.io.Namedlist) -> NamedListJson:
    named: Dict[str, IONamedListItem] = {}
    for k, vs in xs.items():
        if not isinstance(vs, list):
            named[k] = annotated_str_to_json(vs)
            continue

        named[k] = [annotated_str_to_json(v) for v in vs]

    named_values = set()
    for vs in named.values():
        if not isinstance(vs, list):
            vs = [vs]

        for v in vs:
            if isinstance(v, dict):
                v = v["value"]
            named_values.add(v)

    unnamed: List[IONamedListItem] = []
    for vs in xs:
        if not isinstance(vs, list):
            vs = [vs]

        for v in vs:
            obj = annotated_str_to_json(v)

            rendered = obj
            if isinstance(rendered, dict):
                rendered = rendered["value"]
            if rendered in named_values:
                continue

            unnamed.append(obj)

    return {"positional": unnamed, "keyword": named}


class SnakemakeJobTask(PythonAutoContainerTask[Pod]):
    def __init__(
        self,
        wf: SnakemakeWorkflow,
        job: snakemake.jobs.Job,
        inputs: Dict[str, Union[Type[LatchFile], Type[LatchDir]]],
        outputs: Dict[str, Union[Type[LatchFile], Type[LatchDir]]],
        target_file_for_input_param: Dict[str, str],
        target_file_for_output_param: Dict[str, str],
        is_target: bool,
        interface: Interface,
    ):
        name = f"{job.name}_{job.jobid}"

        self.wf = wf
        self.job = job
        self._is_target = is_target
        self._python_inputs = inputs
        self._python_outputs = outputs
        self._target_file_for_input_param = target_file_for_input_param
        self._target_file_for_output_param = target_file_for_output_param

        self._task_function = task_fn_placeholder

        limits = self.job.resources
        cores = limits.get("cpus", 4)
        # convert MB to GiB
        mem = limits.get("mem_mb", 8589) * 1000 * 1000 // 1024 // 1024 // 1024
        gpu = limits.get("nvidia_gpu", 0)

        self._uses_gpu = gpu > 0
        if self._uses_gpu and self.job.container_img_url is not None:
            click.secho(
                dedent("""
                    GPU tasks within container images are not yet supported. To resolve,
                    use conda for all GPU tasks OR add all dependencies to your main
                    Dockerfile.
                    """),
                fg="red",
            )
            raise click.exceptions.Exit(1)

        task_config = None
        if not self._uses_gpu:
            task_config = custom_task(cpu=cores, memory=mem).keywords["task_config"]
        else:
            if gpu == 1 and cores <= 7 and mem <= 30:
                task_config = _get_small_gpu_pod()
            elif gpu == 1 and cores <= 31 and mem <= 120:
                task_config = _get_large_gpu_pod()
            else:
                click.secho(
                    dedent(f"""
                        GPU task resource limit is too high.
                        request: gpu={gpu}, cores={cores}, RAM={mem} GiB
                        max:     gpu=1, cores=31, RAM=120 GiB)",
                        """),
                    fg="red",
                )
                raise click.exceptions.Exit(1)

        super().__init__(
            task_type="sidecar",
            task_type_version=2,
            name=name,
            interface=interface,
            task_config=task_config,
            task_resolver=SnakemakeJobTaskResolver(),
        )

    # todo(maximsmol): this is very awful
    def _serialize_pod_spec(self, settings: SerializationSettings) -> Dict[str, Any]:
        containers = self.task_config.pod_spec.containers
        primary_exists = False
        for container in containers:
            if container.name == self.task_config.primary_container_name:
                primary_exists = True
                break
        if not primary_exists:
            # insert a placeholder primary container if it is not defined in the pod spec.
            containers.append(V1Container(name=self.task_config.primary_container_name))

        final_containers = []
        sdk_default_container = super().get_container(settings)
        for container in containers:
            # In the case of the primary container, we overwrite specific container attributes with the default values
            # used in the regular Python task.
            if container.name == self.task_config.primary_container_name:
                container.image = sdk_default_container.image
                # Spawn entrypoint as child process so it can receive signals
                container.command = [
                    "/bin/bash",
                    "-c",
                    (
                        "exec 3>&1 4>&2 && latch cp"
                        f' "latch:///.snakemake_latch/workflows/{self.wf.name}/{self.wf.jit_wf_version}/{self.wf.jit_exec_display_name}/entrypoint.py"'
                        " latch_entrypoint.py && ("
                        f" {' '.join(sdk_default_container.args)} 1>&3 2>&4 )"
                    ),
                ]
                container.args = []

                limits, requests = {}, {}
                for resource in sdk_default_container.resources.limits:
                    limits[_sanitize_resource_name(resource)] = resource.value
                for resource in sdk_default_container.resources.requests:
                    requests[_sanitize_resource_name(resource)] = resource.value

                resource_requirements = V1ResourceRequirements(
                    limits=limits, requests=requests
                )
                if len(limits) > 0 or len(requests) > 0:
                    # Important! Only copy over resource requirements if they are non-empty.
                    container.resources = resource_requirements

                container.env = [
                    V1EnvVar(name=key, value=val)
                    for key, val in sdk_default_container.env.items()
                ]

            final_containers.append(container)

        self.task_config._pod_spec.containers = final_containers

        return ApiClient().sanitize_for_serialization(self.task_config.pod_spec)

    def get_k8s_pod(self, settings: SerializationSettings) -> _task_models.K8sPod:
        return _task_models.K8sPod(
            pod_spec=self._serialize_pod_spec(settings),
            metadata=_task_models.K8sObjectMetadata(
                labels=self.task_config.labels,
                annotations=self.task_config.annotations,
            ),
        )

    def get_container(self, settings: SerializationSettings) -> _task_models.Container:
        return None

    def get_config(self, settings: SerializationSettings) -> Dict[str, str]:
        return {_PRIMARY_CONTAINER_NAME_FIELD: self.task_config.primary_container_name}

    def get_fn_interface(self):
        res = ""

        params_str = ",\n".join(
            reindent(
                rf"""
                {param}: {t.__name__}
                """,
                1,
            ).rstrip()
            for param, t in self._python_inputs.items()
        )

        outputs_str = "None:"
        if len(self._python_outputs.items()) > 0:
            output_fields = "\n".join(
                reindent(
                    rf"""
                    {param}: {t.__name__}
                    """,
                    1,
                ).rstrip()
                for param, t in self._python_outputs.items()
            )

            res += reindent(
                rf"""
                class Res{self.name}(NamedTuple):
                __output_fields__

                """,
                0,
            ).replace("__output_fields__", output_fields)
            outputs_str = f"Res{self.name}:"

        res += (
            reindent(
                rf"""
                task = custom_task(cpu=-1, memory=-1) # these limits are a lie and are ignored when generating the task spec
                @task(cache=True)
                def {self.name}(
                __params__
                ) -> __outputs__
                """,
                0,
            )
            .replace("__params__", params_str)
            .replace("__outputs__", outputs_str)
        )
        return res

    def get_fn_return_stmt(self, remote_output_url: Optional[str] = None):
        print_outs: List[str] = []
        results: List[str] = []
        for out_name, out_type in self._python_outputs.items():
            target_path = self._target_file_for_output_param[out_name]

            remote_path = None
            if not self._is_target:
                remote_path = f"/.snakemake_latch/workflows/{self.wf.name}/task_outputs/{self.wf.jit_wf_version}/{self.wf.jit_exec_display_name}/{self.name}/{target_path}"
            elif remote_output_url is None:
                remote_path = Path("/Snakemake Outputs") / self.wf.name / target_path
            else:
                remote_path = Path(urlparse(remote_output_url).path) / target_path

            results.append(
                reindent(
                    rf"""
                    {out_name}={out_type.__name__}({repr(target_path)}, "latch://{remote_path}")
                    """,
                    2,
                ).rstrip()
            )

            print_outs.append(
                reindent(
                    rf"""
                    print(f'  {out_name}={{file_name_and_size(Path("{target_path}"))}} -> latch://{remote_path}')
                    """,
                    1,
                )
            )

        print_out_str = "\n".join(print_outs)
        return_str = ",\n".join(results)

        return (
            reindent(
                rf"""
                    print("Uploading results:")
                __print_out__

                    return Res{self.name}(
                __return_str__
                    )
            """,
                0,
            )
            .replace("__print_out__", print_out_str)
            .replace("__return_str__", return_str)
        )

    def get_fn_code(
        self,
        snakefile_path_in_container: str,
        remote_output_url: Optional[str] = None,
        overwrite_config: Optional[Dict[str, str]] = None,
    ):
        code_block = self.get_fn_interface()

        for param, t in self._python_inputs.items():
            if not issubclass(t, (LatchFile, LatchDir)):
                continue

            code_block += reindent(
                rf"""
                {param}_dst_p = Path("{self._target_file_for_input_param[param]}")

                print(f"Downloading {param}: {{{param}.remote_path}}")
                {param}_p = Path({param}).resolve()
                print(f"  {{file_name_and_size({param}_p)}}")

                """,
                1,
            )

            code_block += reindent(
                rf"""
                print(f"Moving {param} to {{{param}_dst_p}}")
                check_exists_and_rename(
                    {param}_p,
                    {param}_dst_p
                )
                """,
                1,
            )

        jobs: List[Job] = [self.job]
        if isinstance(self.job, GroupJob):
            jobs = self.job.jobs

        if overwrite_config is not None:
            for param, val in overwrite_config.items():
                self.job.rule.workflow.globals["config"][param] = val

        container_args = []
        if self.wf._use_container:
            args = self.wf._container_args
            if self._uses_gpu:
                args.append("--gpus all")

            if len(args) > 0:
                container_args = ["--singularity-args"] + args

        snakemake_args = [
            "-m",
            "latch_cli.snakemake.single_task_snakemake",
            "-s",
            snakefile_path_in_container,
            *(["--use-conda"] if self.wf._use_conda else []),
            *(["--use-singularity"] if self.wf._use_container else []),
            # we use docker instead of singularity, so these are docker run arguments
            *container_args,
            "--target-jobs",
            *jobs_cli_args(jobs),
            "--allowed-rules",
            *self.job.rules,
            "--local-groupid",
            str(self.job.jobid),
            "--cores",
            str(self.wf.cores),
        ]
        if not self.job.is_group():
            snakemake_args.append("--force-use-threads")

        excluded = {"_nodes", "_cores", "tmpdir"}
        allowed_resources = list(
            filter(lambda x: x[0] not in excluded, self.job.resources.items())
        )
        if len(allowed_resources) > 0:
            snakemake_args.append("--resources")
            for resource, value in allowed_resources:
                snakemake_args.append(f"{resource}={value}")

        snakemake_data = {
            "rules": {},
            "outputs": self.job.output,
            "overwrite_config": overwrite_config,
        }

        for job in jobs:
            snakemake_data["rules"][job.rule.name] = {
                "inputs": named_list_to_json(job.input),
                "outputs": named_list_to_json(job.output),
                "params": {
                    "keyword": {k: str(v) for k, v in job.params.items()},
                    "positional": [],
                },
                "benchmark": job.benchmark,
                "log": job.log,
                "shellcmd": job.shellcmd,
            }

        if remote_output_url is None:
            remote_path = Path("/Snakemake Outputs") / self.wf.name / self.job.name
        else:
            remote_path = Path(urlparse(remote_output_url).path)

        log_files = self.job.log if self.job.log is not None else []
        # for debugging purposes, we upload outputs to LData even if
        # they are not target files or used in a downstream rule
        result_files = set([
            self._target_file_for_output_param[out_name]
            for out_name, _ in self._python_outputs.items()
        ])
        unused_outputs = (
            [out for out in self.job.output if out not in result_files]
            if self.job.output is not None
            else []
        )

        code_block += reindent(
            rf"""
            lp = LatchPersistence()
            compiled = Path("compiled.py")
            print("Saving compiled Snakemake script")
            with compiled.open("w") as f:
                try:
                    subprocess.run(
                        [sys.executable,{','.join(repr(x) for x in [*snakemake_args, "--print-compilation"])}],
                        check=True,
                        env={{
                            **os.environ,
                            "LATCH_SNAKEMAKE_DATA": {repr(json.dumps(snakemake_data))},
                            "LATCH_PRINT_COMPILATION": "1"
                        }},
                        stdout=f
                    )
                except CalledProcessError:
                    print("  Failed")
                except Exception:
                    traceback.print_exc()
            lp.upload(compiled, "latch:///.snakemake_latch/workflows/{self.wf.name}/compiled_tasks/{self.name}.py")
            """,
            1,
        )

        if (
            self.wf._docker_metadata is not None
            and self.job.container_img_url is not None
        ):
            code_block += reindent(
                rf"""

                print("\n\n\nLogging into Docker\n")
                from latch.functions.secrets import get_secret
                docker_usr = "{self.wf._docker_metadata.username}"
                try:
                    docker_pwd = get_secret("{self.wf._docker_metadata.secret_name}")
                except ValueError as e:
                    print("Failed to get Docker credentials:", e)
                    sys.exit(1)

                try:
                    subprocess.run(
                        [
                            "docker",
                            "login",
                            "--username",
                            docker_usr,
                            "--password",
                            docker_pwd,
                        ],
                        check=True,
                    )
                except CalledProcessError as e:
                    print("Failed to login to Docker")
                except Exception:
                    traceback.print_exc()
                """,
                1,
            )

        code_block += reindent(
            rf"""

            print("\n\n\nRunning snakemake task\n")
            try:
                log_files = {repr(log_files)}
                try:
                    tail = None
                    if len(log_files) == 1:
                        log = Path(log_files[0])
                        log.parent.mkdir(parents=True, exist_ok=True)
                        log.touch()

                        print(f"Tailing the only log file: {{log}}")
                        tail = subprocess.Popen(["tail", "--follow", log])

                    print("\n\n\n")
                    try:
                        subprocess.run(
                            [sys.executable,{','.join(repr(x) for x in snakemake_args)}],
                            check=True,
                            env={{
                                **os.environ,
                                "LATCH_SNAKEMAKE_DATA": {repr(json.dumps(snakemake_data))}
                            }}
                        )
                    finally:
                        if tail is not None:
                            import signal
                            tail.send_signal(signal.SIGINT)
                            try:
                                tail.wait(1)
                            except subprocess.TimeoutExpired:
                                tail.kill()

                            tail.wait()
                            # -2 is SIGINT
                            if tail.returncode != -2 and tail.returncode != 0:
                                print(f"\n\n\n[!] Log file tail died with code {{tail.returncode}}")

                    print("\n\n\nDone\n\n\n")
                except Exception as e:
                    print("\n\n\n[!] Failed\n\n\n")
                    raise e
                finally:
                    print("Uploading logs:")
                    for x in log_files:
                        local = Path(x)
                        remote = f"latch://{remote_path}/{{str(local).removeprefix('/')}}"

                        if not local.exists():
                            print("  Does not exist")
                            continue

                        print(f"  {{file_name_and_size(local)}} -> {{remote}}")

                        if local.is_file():
                            lp.upload(local, remote)
                        else:
                            lp.upload_directory(str(local), remote)

                        print("    Done")

                    print("Uploading intermediate outputs:")
                    for x in {repr(unused_outputs)}:
                        local = Path(x)
                        remote = f"latch://{remote_path}/{{str(local).removeprefix('/')}}"

                        if not local.exists():
                            print(f"  {{x}}: Does not exist")
                            continue

                        print(f"  {{file_name_and_size(local)}} -> {{remote}}")

                        if local.is_file():
                            lp.upload(local, remote)
                        else:
                            lp.upload_directory(str(local), remote)

                        print("    Done")

                    benchmark_file = {repr(self.job.benchmark)}
                    if benchmark_file is not None:
                        print("\nUploading benchmark:")

                        local = Path(benchmark_file)
                        if local.exists():
                            print(local.read_text())

                            remote = f"latch://{remote_path}/{{str(local).removeprefix('/')}}"
                            print(f"  {{file_name_and_size(local)}} -> {{remote}}")
                            lp.upload(local, remote)
                            print("    Done")
                        else:
                            print("  Does not exist")

            finally:
                ignored_paths = {{".cache", ".snakemake/conda"}}
                ignored_names = {{".git", ".latch", "__pycache__"}}
            """,
            1,
        )

        code_block += self.get_fn_return_stmt(remote_output_url=remote_output_url)
        return code_block

    @property
    def dockerfile_path(self) -> Path:
        return self._dockerfile_path

    @property
    def task_function(self):
        return self._task_function

    def execute(self, **kwargs) -> Any:
        return exception_scopes.user_entry_point(self._task_function)(**kwargs)


class SnakemakeJobTaskResolver(DefaultTaskResolver):
    @property
    def location(self) -> str:
        return "flytekit.core.python_auto_container.default_task_resolver"

    def loader_args(
        self, settings: SerializationSettings, task: SnakemakeJobTask
    ) -> List[str]:
        return ["task-module", "latch_entrypoint", "task-name", task.name]

    def load_task(self, loader_args: List[str]) -> PythonAutoContainerTask:
        _, task_module, _, task_name, *_ = loader_args

        task_module = importlib.import_module(task_module)

        task_def = getattr(task_module, task_name)
        return task_def


class JITRegisterWorkflowResolver(DefaultTaskResolver):
    @property
    def location(self) -> str:
        return "flytekit.core.python_auto_container.default_task_resolver"

    def loader_args(
        self, settings: SerializationSettings, task: PythonAutoContainerTask[T]
    ) -> List[str]:
        return ["task-module", "snakemake_jit_entrypoint", "task-name", task.name]

    def load_task(self, loader_args: List[str]) -> PythonAutoContainerTask:
        _, task_module, _, task_name, *_ = loader_args

        task_module = importlib.import_module(task_module)

        task_def = getattr(task_module, task_name)
        return task_def



from latch.resources.tasks import small_task
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile


@small_task
def task(input_file: LatchFile, output_directory: LatchOutputDir) -> LatchFile:
    raise NotImplementedError("Task not implemented.")



import subprocess

from latch.resources.tasks import small_task
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile


# change the name of this function to something more descriptive
@small_task
def r_task(input_file: LatchFile, output_directory: LatchOutputDir) -> LatchFile:
    # You can run R files as a subprocess:

    subprocess.run(
        " ".join(
            [
                "Rscript",
                "path/to/your_script.R",
                "command_line_arg_1",
                "command_line_arg_2",
            ]
        ),
        check=True,
        shell=True,
    )
    ...
    output_location = f"{output_directory.remote_directory}/YOUR_OUTPUT_FILE_NAME"
    return LatchFile(str("/root/LOCAL_FILE_IN_TASK"), output_location)



import subprocess
from pathlib import Path
from typing import List

from latch.functions.messages import message
from latch.resources.tasks import small_task
from latch.types.directory import LatchOutputDir
from latch.types.file import LatchFile


def run(cmd: List[str]):
    try:
        return subprocess.run(cmd, check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        stderr = e.stderr.decode("utf-8")
        if stderr != "":
            message(
                "error",
                {"title": "Blastp Failed", "body": f"Stderr: {stderr}"},
            )
        print(stderr)
        raise e


@small_task
def blastp_task(
    query_fasta: LatchFile, database: LatchFile, output_directory: LatchOutputDir
) -> LatchFile:
    outdir = Path("/root/outputs").resolve()
    outdir.mkdir(exist_ok=True)

    database_path = Path(database.local_path)
    query_file_path = Path(query_fasta.local_path)

    db_prep = [
        "docker",
        "run",
        "--user",
        "root",  # run as root to avoid permission issues
        "-v",
        f"{database_path.parent}:/data/",  # mount the database file directory into the container
        "biocontainers/blast:2.2.31",
        "makeblastdb",
        "-in",
        f"/data/{database_path.name}",
        "-dbtype",
        "prot",
    ]

    db_prep_run = run(db_prep)
    print(db_prep_run.stdout.decode("utf-8"))

    # run blast
    blast_cmd = [
        "docker",
        "run",
        "--user",  # run as root to avoid permission issues
        "root",
        "-v",
        f"{query_file_path.parent}:/data/",  # mount the query file directory into the container
        "-v",
        f"{database_path.parent}:/db/",  # mount the database file directory into the container
        "-v",
        f"{outdir}:/output/",  # mount the output directory into the container
        "biocontainers/blast:2.2.31",
        "blastp",
        "-query",
        f"/data/{Path(query_fasta.local_path).name}",
        "-db",
        f"/db/{database_path.name}",
        "-out",
        f"/output/results.txt",
    ]

    blast_run = run(blast_cmd)
    print(blast_run.stdout.decode("utf-8"))

    # intended output path of the file in Latch console, constructed from
    # the user provided output directory
    output_location = (
        f"{output_directory.remote_directory}results.txt"
        if output_directory.remote_directory.endswith("/")
        else f"{output_directory.remote_directory}/results.txt"
    )
    local_sam_file = outdir / "results.txt"

    return LatchFile(str(local_sam_file), output_location)



import hashlib
import secrets
from typing import Tuple

from latch_cli.auth.utils import _base64url_encode


class PKCE:
    """A Context manager to control state for PKCE flow.

    The Proof Key for Code Exchange is outlined rigorously in `RFC7636`_.  The
    implementation in this module was written using that document as gospel. It
    is an extension of OAuth2.0 that adds crytographic secrets to prevent
    request interception.

    Here is a diagram to summarize the flow (capital letters correspond to
    descriptions):

    ::

                                                      +-------------------+
                                                      |   Authz Server    |
            +--------+                                | +---------------+ |
            |        |--(A)- Authorization Request ---->|               | |
            |        |       + t(code_verifier), t_m  | | Authorization | |
            |        |                                | |    Endpoint   | |
            |        |<-(B)---- Authorization Code -----|               | |
            |        |                                | +---------------+ |
            | Client |                                |                   |
            |        |                                | +---------------+ |
            |        |--(C)-- Access Token Request ---->|               | |
            |        |          + code_verifier       | |    Token      | |
            |        |                                | |   Endpoint    | |
            |        |<-(D)------ Access Token ---------|               | |
            +--------+                                | +---------------+ |
                                                      +-------------------+

                          Figure 2: Abstract Protocol Flow


    A. The client creates and records a secret named the "code_verifier"
       and derives a transformed version "t(code_verifier)" (referred to
       as the "code_challenge"), which is sent in the OAuth 2.0
       Authorization Request along with the transformation method "t_m".

    B. The Authorization Endpoint responds as usual but records
       "t(code_verifier)" and the transformation method.

    C. The client then sends the authorization code in the Access Token
       Request as usual but includes the "code_verifier" secret generated
       at (A).

    D. The authorization server transforms "code_verifier" and compares
       it to "t(code_verifier)" from (B).  Access is denied if they are
       not equal.

    An attacker who intercepts the authorization code at (B) is unable to
    redeem it for an access token, as they are not in possession of the
    "code_verifier" secret.

    Example usage: ::

        with PKCE() as pkce:
            oauth2_flow = OAuth2(pkce, ..)


    .. _RFC7636:
        https://datatracker.ietf.org/doc/html/rfc7636
    """

    challenge_method = "S256"
    """The challenge method used to encode the code verifier.

    'If the client is capable of using "S256", it MUST use "S256", as
    "S256" is Mandatory To Implement (MTI) on the server.'

    .. _RFC7636#section-4.2:
        https://datatracker.ietf.org/doc/html/rfc7636#section-4.2

    Thus this value is hardcoded.
    """

    def __init__(self):
        self.verifier, self.challenge = self.construct_challenge()

    def __enter__(self, *args):
        return self

    def __exit__(self, *args): ...

    def construct_challenge(self) -> Tuple[str, str]:
        """Construct verifier & challenge to verify a client's identity in PKCE.

        Reference  `RFC7636`_.

        Returns:
            code verifier: A cryptographically random string that is used to
                correlate the authorization request to the token request.
            code challenge: A challenge derived from the code verifier that is
                sent in the authorization request, to be verified against later.

        .. _RFC7636:
            https://datatracker.ietf.org/doc/html/rfc7636
        """
        verifier = _base64url_encode(secrets.token_bytes(32))
        challenge = _base64url_encode(hashlib.sha256(verifier.encode()).digest())
        return verifier, challenge



import click

from latch.ldata.path import LatchPathError, LPath


def mkdirp(remote_directory):
    """Creates an empty directory on Latch

    Args:
        remote_directory:   A valid path to a remote destination, of the form

                                [latch://] [/] dir_1/dir_2/.../dir_n/dir_name,

                            where dir_name is the name of the new directory to be created.
                            Every directory in the path (dir_i) must already exist.

    This function will create a directory at the specified path in Latch. Will error if
    the path is invalid or if an upstream directory does not exist. If a directory with the
    same name already exists, this will make a new directory with an indexed name (see below).

    Example: ::

        mkdirp("sample") # sample doesn't already exist

            Creates a new empty directory visible in Latch Console called sample, located in
            the root of the user's Latch filesystem

        mkdirp("/dir1/doesnt_exist/dir2/") # doesnt_exist doesn't exist

            Creates a two new directories visible in Latch Console called "doesnt_exist" and
            "dir2", located in the directory dir1.
    """
    try:
        LPath(remote_directory).mkdirp()
    except LatchPathError as e:
        click.secho(str(e), fg="red")
        raise click.exceptions.Exit(1) from e
    click.secho(f"Successfully created {remote_directory}.", fg="green")



from os import PathLike
from pathlib import Path
from typing import List, Optional, Type, TypedDict, Union, get_args, get_origin
from urllib.parse import urlparse

import gql
from flytekit.core.annotation import FlyteAnnotation
from flytekit.core.context_manager import FlyteContext, FlyteContextManager
from flytekit.core.type_engine import TypeEngine, TypeTransformer
from flytekit.exceptions.user import FlyteUserException
from flytekit.models.literals import Literal
from flytekit.types.directory.types import (
    FlyteDirectory,
    FlyteDirToMultipartBlobTransformer,
)
from latch_sdk_gql.execute import execute
from typing_extensions import Annotated

from latch.types.file import LatchFile
from latch.types.utils import format_path, is_valid_url
from latch_cli.utils import urljoins
from latch_cli.utils.path import normalize_path


class IterdirChild(TypedDict):
    type: str
    name: str


class IterdirChildLdataTreeEdge(TypedDict):
    child: IterdirChild


class IterdirChildLdataTreeEdges(TypedDict):
    nodes: List[IterdirChildLdataTreeEdge]


class IterDirLDataResolvePathFinalLinkTarget(TypedDict):
    childLdataTreeEdges: IterdirChildLdataTreeEdges


class IterdirLdataResolvePathData(TypedDict):
    finalLinkTarget: IterDirLDataResolvePathFinalLinkTarget


class NodeDescendantsNode(TypedDict):
    relPath: str


class NodeDescendantsDescendants(TypedDict):
    nodes: List[NodeDescendantsNode]


class NodeDescendantsFinalLinkTarget(TypedDict):
    descendants: NodeDescendantsDescendants


class NodeDescendantsLDataResolvePathData(TypedDict):
    finalLinkTarget: NodeDescendantsFinalLinkTarget


class LatchDir(FlyteDirectory):
    """Represents a directory in the context of a task execution.

    The local path identifies the directory location on local disk in
    the context of a task execution.

    ..
        @task
        def task(dir: LatchDir):

            mypath = Path(dir).joinpath("foo.txt").resolve()


    The remote path identifies a remote location. The remote location, either a
    latch or s3 url, can be inspected from an object passed to the task to
    reveal its remote source.

    It can also to deposit the file to a latch path when the directory is
    returned from a task.

    ..

        @task
        def task(dir: LatchFile):


            # Manipulate directory contents locally and return back to
            # LatchData

            return LatchDir("./foo", "latch:///foo")
    """

    def __init__(
        self,
        path: Union[str, PathLike],
        remote_path: Optional[PathLike] = None,
        **kwargs,
    ):
        if path is None:
            raise ValueError("Unable to instantiate LatchDir with None")

        # Cast PathLike objects so that LatchDir has consistent JSON
        # representation.
        parsed = urlparse(str(path))
        if parsed.scheme == "file":
            self.path = parsed.path
        elif parsed.scheme == "latch":
            self.path = normalize_path(str(path))
        else:
            self.path = str(path)

        self._path_generated = False

        if is_valid_url(self.path) and remote_path is None:
            self._remote_directory = self.path
        else:
            self._remote_directory = None if remote_path is None else str(remote_path)

        if kwargs.get("downloader") is not None:
            super().__init__(self.path, kwargs["downloader"], self._remote_directory)
        else:

            def downloader():
                ctx = FlyteContextManager.current_context()
                if (
                    ctx is not None
                    and hasattr(self, "_remote_directory")
                    and self._remote_directory is not None
                    # todo(kenny) is this necessary?
                    and ctx.inspect_objects_only is False
                ):
                    self._idempotent_set_path()

                    return ctx.file_access.get_data(
                        self._remote_directory,
                        self.path,
                        is_multipart=True,
                    )

            super().__init__(self.path, downloader, self._remote_directory)

    def _idempotent_set_path(self):
        if self._path_generated:
            return

        ctx = FlyteContextManager.current_context()
        if ctx is None:
            return

        self.path = ctx.file_access.get_random_local_directory()
        self._path_generated = True

    def iterdir(self) -> List[Union[LatchFile, "LatchDir"]]:
        ret: List[Union[LatchFile, "LatchDir"]] = []

        if self.remote_path is None:
            for child in Path(self.path).iterdir():
                if child.is_dir():
                    ret.append(LatchDir(str(child)))
                else:
                    ret.append(LatchFile(str(child)))

            return ret

        res: Optional[IterdirLdataResolvePathData] = execute(
            gql.gql("""
            query LDataChildren($argPath: String!) {
                ldataResolvePathData(argPath: $argPath) {
                    finalLinkTarget {
                        childLdataTreeEdges(filter: { child: { removed: { equalTo: false } } }) {
                            nodes {
                                child {
                                    name
                                    type
                                }
                            }
                        }
                    }
                }
            }"""),
            {"argPath": self.remote_path},
        )["ldataResolvePathData"]

        if res is None:
            raise ValueError(f"No directory found at path: {self}")

        for node in res["finalLinkTarget"]["childLdataTreeEdges"]["nodes"]:
            child = node["child"]

            path = urljoins(self.remote_path, child["name"])
            if child["type"] == "DIR":
                ret.append(LatchDir(path))
            else:
                ret.append(LatchFile(path))

        return ret

    def _create_imposters(self):
        self._idempotent_set_path()

        res: Optional[NodeDescendantsLDataResolvePathData] = execute(
            gql.gql("""
                query NodeDescendantsQuery($path: String!) {
                    ldataResolvePathData(argPath: $path) {
                        finalLinkTarget {
                            descendants {
                                nodes {
                                    relPath
                                }
                            }
                        }
                    }
                }
            """),
            {"path": self._remote_directory},
        )["ldataResolvePathData"]

        if res is None:
            # todo(ayush): proper error message + exit
            raise FlyteUserException(f"No directory at {self._remote_directory}")

        root = Path(self.path)
        for x in res["finalLinkTarget"]["descendants"]["nodes"]:
            p = root / x["relPath"]

            p.parent.mkdir(exist_ok=True, parents=True)
            p.touch(exist_ok=True)

    @property
    def local_path(self) -> str:
        """File path local to the environment executing the task."""

        # This will manually download object to local disk in the case of a
        # user wishing to access the file without using `open`.
        self.__fspath__()

        return str(self.path)

    @property
    def remote_path(self) -> Optional[str]:
        """A url referencing in object in LatchData or s3."""
        return self._remote_directory

    def __repr__(self):
        if self.remote_path is None:
            return f"LatchDir({repr(format_path(self.local_path))})"

        return (
            f"LatchDir({repr(self.path)},"
            f" remote_path={repr( format_path(self.remote_path))})"
        )

    def __str__(self):
        if self.remote_path is None:
            return "LatchDir()"

        return f"LatchDir({format_path(self.remote_path)})"


LatchOutputDir = Annotated[
    LatchDir,
    FlyteAnnotation(
        {"output": True},
    ),
]
"""A LatchDir tagged as the output of some workflow.

The Latch Console uses this metadata to avoid checking for existence of the
file at its remote path and displaying an error. This check is normally made to
avoid launching workflows with LatchDirs that point to objects that don't
exist.
"""


class LatchDirPathTransformer(FlyteDirToMultipartBlobTransformer):
    def __init__(self):
        TypeTransformer.__init__(self, name="LatchDirPath", t=LatchDir)

    def to_python_value(
        self,
        ctx: FlyteContext,
        lv: Literal,
        expected_python_type: Union[Type[LatchDir], PathLike],
    ) -> FlyteDirectory:
        uri = lv.scalar.blob.uri
        if expected_python_type is PathLike:
            raise TypeError(
                "Casting from Pathlike to LatchDir is currently not supported."
            )

        while get_origin(expected_python_type) == Annotated:
            expected_python_type = get_args(expected_python_type)[0]

        if not issubclass(expected_python_type, LatchDir):
            raise TypeError(
                f"Neither os.PathLike nor LatchDir specified {expected_python_type}"
            )

        # This is a local file path, like /usr/local/my_file, don't mess with it. Certainly, downloading it doesn't
        # make any sense.
        if not ctx.file_access.is_remote(uri):
            return expected_python_type(uri)

        # For the remote case, return an FlyteDirectory object that can download
        local_folder = ctx.file_access.get_random_local_directory()

        def _downloader():
            return ctx.file_access.get_data(uri, local_folder, is_multipart=True)

        ret = LatchDir(local_folder, uri, downloader=_downloader)
        ret._remote_source = uri
        return ret


TypeEngine.register(LatchDirPathTransformer())



import os
import sys
import termios
import textwrap
import tty
from pathlib import Path
from typing import Dict, List

from apscheduler.schedulers.background import BackgroundScheduler
from latch_sdk_config.latch import config

import latch_cli.menus as menus
from latch_cli.tinyrequests import post
from latch_cli.utils import account_id_from_token, current_workspace, retrieve_or_login


def get_executions():
    """Open an interactive terminal user interface that shows all executions for
    a particular user on Latch.

    This function in many ways mimics the interface at https://console.latch.bio
    but in a terminal. You may scroll around using either the arrow keys (or
    [HJKL] if you like Vim) and you can hit [ENTER] to see more information
    about the selected execution. You can see logs, and abort a running
    execution.

    This function should only be called from the CLI, as doing so in any other
    setting will likely cause the interface to behave in an unexpected way.
    """

    token = retrieve_or_login()
    context = current_workspace()
    headers = {"Authorization": f"Bearer {token}"}

    resp = post(
        url=config.api.execution.list,
        headers=headers,
        json={"ws_account_id": context},
    )

    resp.raise_for_status()

    options = []
    data: dict[str, dict] = resp.json()
    display_columns = [
        "display_name",
        "workflow_tagged",
        "start_time",
        "status",
    ]

    for execution_data in sorted(data.values(), key=lambda x: -int(x["id"])):
        options.append(
            {
                **execution_data,
                "workflow_tagged": f'{execution_data["workflow_name"]}/{execution_data["workflow_version"]}',
            }
        )

    _all_executions(
        title="All Executions",
        column_names=display_columns,
        options=options,
    )


def _all_executions(
    title: str,
    column_names: List[str],
    options: List[Dict[str, str]],
):
    if len(options) == 0:
        raise ValueError("No executions to show.")
    elif len(column_names) == 0:
        raise ValueError("No column names specified")

    term_width, term_height = os.get_terminal_size()

    def render(
        curr_selected: int,
        hor_index: int,
        term_width: int,
        term_height: int,
    ) -> int:
        # DISCLAIMER : MOST OF THE MAGIC NUMBERS HERE WERE THROUGH TRIAL AND ERROR

        menus.move_cursor((2, 2))

        max_per_page = term_height - 5
        vert_index = max(0, curr_selected - max_per_page // 2)

        lengths = {col: len(col) for col in column_names}
        for j in range(len(options)):
            values = options[j]
            for col in column_names:
                if values[col] is None:
                    values[col] = ""
                lengths[col] = max(lengths[col], len(values[col]))

        if len(column_names) > 1:
            column_spacing = max(
                2,
                (term_width - 7 - sum(lengths.values())) // (len(column_names) - 1),
            )
        else:
            column_spacing = 2

        max_row_len = sum(lengths.values()) + column_spacing * (len(column_names) - 1)

        menus._print(title)
        menus.line_down(2)

        for i in range(vert_index, vert_index + max_per_page):
            if i >= len(options):
                break
            values = options[i]
            row_str = ""
            for j, col in enumerate(column_names):
                item = values[col]
                if j == len(column_names) - 1:
                    row_str = row_str + f"{item: <{lengths[col]}}"
                else:
                    row_str = row_str + f"{item: <{lengths[col] + column_spacing}}"

            row_str = row_str[hor_index : hor_index + term_width - 6]

            if i == curr_selected:
                green = "\x1b[38;5;40m"
                bold = "\x1b[1m"
                reset = "\x1b[0m"
                row_str = f"{green}{bold}{row_str}{reset}"

            menus.move_cursor_right(3)
            menus._print(row_str)
            menus.line_down(1)

        menus.move_cursor((2, term_height - 1))
        control_str = "[ARROW-KEYS] Navigate\t[ENTER] Select\t[Q] Quit"
        menus._print(control_str)
        menus.draw_box((2, 3), term_height - 5, term_width - 4)

        menus._show()
        return max_row_len

    old_settings = termios.tcgetattr(sys.stdin.fileno())
    tty.setraw(sys.stdin.fileno())

    curr_selected = hor_index = 0

    menus.remove_cursor()
    menus.clear_screen()
    menus.move_cursor((0, 0))

    prev = (curr_selected, hor_index, term_width, term_height)
    max_row_len = render(curr_selected, hor_index, term_width, term_height)

    try:
        while True:
            b = menus.read_bytes(1)
            term_width, term_height = os.get_terminal_size()
            rerender = False
            if b == b"\r":
                selected_execution_data = options[curr_selected]
                resp = post(
                    config.api.workflow.graph,
                    headers={"Authorization": f"Bearer {retrieve_or_login()}"},
                    json={
                        "workflow_id": selected_execution_data["workflow_id"],
                        "execution_id": selected_execution_data["id"],
                    },
                )
                _execution_dashboard(selected_execution_data, resp.json())
                rerender = True
            elif b == b"\x1b":
                b = menus.read_bytes(2)
                if b == b"[A":  # Up Arrow
                    curr_selected = max(0, curr_selected - 1)
                elif b == b"[B":  # Down Arrow
                    curr_selected = min(len(options) - 1, curr_selected + 1)
                elif b == b"[D":  # Left Arrow
                    if max_row_len > term_width + 7:
                        hor_index = max(0, hor_index - 5)
                elif b == b"[C":  # Right Arrow
                    if max_row_len > term_width + 7:
                        hor_index = min(max_row_len - term_width + 7, hor_index + 5)
                elif b == b"[1":  # Start of SHIFT + arrow keys
                    b = menus.read_bytes(3)
                    if b == b";2A":  # Up Arrow
                        curr_selected = max(0, curr_selected - 20)
                    elif b == b";2B":  # Down Arrow
                        curr_selected = min(len(options) - 1, curr_selected + 20)
                    elif b == b";2D":  # Left Arrow
                        if max_row_len > term_width + 7:
                            hor_index = max(0, hor_index - 25)
                    elif b == b";2C":  # Right Arrow
                        if max_row_len > term_width + 7:
                            hor_index = min(
                                max_row_len - term_width + 7, hor_index + 25
                            )
            elif b == b"k":
                curr_selected = max(curr_selected - 1, 0)
            elif b == b"j":
                curr_selected = min(curr_selected + 1, len(options) - 1)
            elif b == b"J":
                curr_selected = min(curr_selected + 20, len(options) - 1)
            elif b == b"K":
                curr_selected = max(curr_selected - 20, 0)
            elif b == b"h":
                if max_row_len > term_width + 7:
                    hor_index = max(0, hor_index - 5)
            elif b == b"l":
                if max_row_len > term_width + 7:
                    hor_index = min(max_row_len - term_width + 7, hor_index + 5)
            elif b == b"H":
                if max_row_len > term_width + 7:
                    hor_index = max(0, hor_index - 25)
            elif b == b"L":
                if max_row_len > term_width + 7:
                    hor_index = min(max_row_len - term_width + 7, hor_index + 25)
            if rerender or (
                (curr_selected, hor_index, term_width, term_height) != prev
            ):
                prev = (curr_selected, hor_index, term_width, term_height)
                menus.clear_screen()
                max_row_len = render(curr_selected, hor_index, term_width, term_height)
    except KeyboardInterrupt: ...
    finally:
        menus.clear_screen()
        menus.reveal_cursor()
        menus.move_cursor((0, 0))
        menus._show()
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, old_settings)


def _execution_dashboard(execution_data: Dict[str, str], workflow_graph: Dict):
    fixed_workflow_graph = list(workflow_graph.items())

    def render(curr_selected: int, term_width: int, term_height: int):
        # DISCLAIMER : MOST OF THE MAGIC NUMBERS HERE WERE THROUGH TRIAL AND ERROR
        menus.move_cursor((2, 2))
        menus._print(
            f'{execution_data["display_name"]} - {execution_data["workflow_tagged"]}'
        )
        menus.draw_box((2, 3), term_height - 5, term_width - 4)

        menus.move_cursor((2, term_height - 1))
        instructions = [
            "[ARROW-KEYS] Navigate",
            "[ENTER] View Task Logs",
            # "[R] Relaunch",
        ]
        if execution_data["status"] == "RUNNING":
            instructions.append("[A] Abort")
        instructions.append("[Q] Back to All Executions")
        menus._print("\t".join(instructions))

        menus.move_cursor((4, 4))
        for i, (_, task) in enumerate(fixed_workflow_graph):
            name, status = task["name"] or task["sub_wf_name"], task["status"]
            row_str = "  ".join([name, status])
            if i == curr_selected:
                green = "\x1b[38;5;40m"
                bold = "\x1b[1m"
                reset = "\x1b[0m"
                menus._print(f"{green}{bold}{row_str}{reset}")
            else:
                menus._print(row_str)
            menus.line_down(1)
            menus.move_cursor_right(3)
        menus._show()

    menus.clear_screen()

    try:
        term_width, term_height = os.get_terminal_size()
        curr_selected = 0

        render(curr_selected, term_width, term_height)
        prev = (curr_selected, term_width, term_height)
        while True:
            b = menus.read_bytes(1)
            rerender = False
            if b == b"\r":
                _log_window(execution_data, fixed_workflow_graph, curr_selected)
                rerender = True
            # elif b in (b"r", b"R"):
            #     relaunch_modal(execution_data)
            #     rerender = True
            elif b in (b"a", b"A"):
                _abort_modal(execution_data)
                rerender = True
            elif b == b"\x1b":
                b = menus.read_bytes(2)
                if b == b"[A":  # Up Arrow
                    curr_selected = max(curr_selected - 1, 0)
                elif b == b"[B":  # Down Arrow
                    curr_selected = min(curr_selected + 1, len(workflow_graph) - 1)
                elif b == b"[1":  # Start of SHIFT + arrow keys
                    b = menus.read_bytes(3)
                    if b == b";2A":  # Up Arrow
                        curr_selected = max(0, curr_selected - 20)
                    elif b == b";2B":  # Down Arrow
                        curr_selected += min(
                            curr_selected + 20, len(workflow_graph) - 1
                        )
            elif b == b"j":
                curr_selected = min(curr_selected + 1, len(workflow_graph) - 1)
            elif b == b"k":
                curr_selected = max(curr_selected - 1, 0)
            elif b == b"J":
                curr_selected = min(curr_selected + 20, len(workflow_graph) - 1)
            elif b == b"K":
                curr_selected = max(curr_selected - 20, 0)
            term_width, term_height = os.get_terminal_size()
            if rerender or (prev != (curr_selected, term_width, term_height)):
                menus.clear_screen()
                prev = (curr_selected, term_width, term_height)
                render(curr_selected, term_width, term_height)
    except KeyboardInterrupt: ...
    finally:
        menus.clear_screen()
        menus.move_cursor((0, 0))
        menus._show()


def _loading_screen(text: str):
    # DISCLAIMER : MOST OF THE MAGIC NUMBERS HERE WERE THROUGH TRIAL AND ERROR
    term_width, term_height = os.get_terminal_size()

    menus.clear_screen()
    menus.draw_box((2, 3), term_height - 5, term_width - 4)

    x = (term_width - len(text)) // 2
    y = term_height // 2

    menus.move_cursor((x, y))
    menus._print(text)
    menus._show()


def _log_window(execution_data, fixed_workflow_graph: list, selected: int):
    # DISCLAIMER : MOST OF THE MAGIC NUMBERS HERE WERE THROUGH TRIAL AND ERROR
    _loading_screen("Loading logs...")

    _, selected_task = fixed_workflow_graph[selected]

    log_file = Path(".latch_execution_log").resolve()
    log_file.touch(exist_ok=True)

    ws_id = current_workspace()
    if ws_id is None or ws_id == "":
        ws_id = account_id_from_token(retrieve_or_login())

    def get_logs():
        resp = post(
            config.api.execution.logs,
            headers={"Authorization": f"Bearer {retrieve_or_login()}"},
            json={
                "exec_id": execution_data["id"],
                "node_id": selected_task["node_id"],
                "account_id": ws_id,
            },
        )
        resp.raise_for_status()
        with open(log_file, "w") as f:
            f.write(resp.json()["message"].replace("\t", "    "))

    def render(vert_index, hor_index, term_width, term_height):
        menus.move_cursor((2, 2))
        menus._print(
            f'{execution_data["display_name"]} - {execution_data["workflow_tagged"]} -'
            f' {selected_task["name"]}'
        )
        menus.draw_box((2, 3), term_height - 5, term_width - 4)
        menus.move_cursor((4, 4))
        with open(log_file, "r") as f:
            for i, line in enumerate(f):
                if i < vert_index:
                    continue
                elif i > vert_index + term_height - 7:
                    continue
                line = line.strip("\n\r")
                menus._print(line[hor_index : hor_index + term_width - 7])
                menus.line_down(1)
                menus.move_cursor_right(3)
        menus.move_cursor((2, term_height - 1))
        control_str = "[ARROW-KEYS] Navigate\t[Q] Back"
        menus._print(control_str)
        menus._show()

    try:
        term_width, term_height = os.get_terminal_size()
        menus.clear_screen()
        get_logs()

        log_sched = BackgroundScheduler()
        log_sched.add_job(
            get_logs,
            "interval",
            seconds=15,
        )
        log_sched.start()

        vert_index = hor_index = 0
        render(vert_index, hor_index, term_width, term_height)
        prev_term_dims = (vert_index, hor_index, term_width, term_height)
        while True:
            b = menus.read_bytes(1)
            rerender = False
            if b in (b"r", b"R"):
                rerender = True
            elif b == b"\x1b":
                b = menus.read_bytes(2)
                if b == b"[A":  # Up Arrow
                    vert_index = max(0, vert_index - 1)
                elif b == b"[B":  # Down Arrow
                    vert_index += 1
                elif b == b"[D":  # Left Arrow
                    hor_index = max(0, hor_index - 5)
                elif b == b"[C":  # Right Arrow
                    hor_index += 5
                elif b == b"[1":  # Start of SHIFT + arrow keys
                    b = menus.read_bytes(3)
                    if b == b";2A":  # Up Arrow
                        vert_index = max(0, vert_index - 20)
                    elif b == b";2B":  # Down Arrow
                        vert_index += 20
                    elif b == b";2D":  # Left Arrow
                        hor_index = max(0, hor_index - 25)
                    elif b == b";2C":  # Right Arrow
                        hor_index += 25
            elif b == b"k":
                vert_index = max(0, vert_index - 1)
            elif b == b"j":
                vert_index += 1
            elif b == b"h":
                hor_index = max(0, hor_index - 5)
            elif b == b"l":
                hor_index += 5
            elif b == b"K":
                vert_index = max(0, vert_index - 20)
            elif b == b"J":
                vert_index += 20
            elif b == b"H":
                hor_index = max(0, hor_index - 25)
            elif b == b"L":
                hor_index += 25
            term_width, term_height = os.get_terminal_size()
            if rerender or (
                prev_term_dims != (vert_index, hor_index, term_width, term_height)
            ):
                menus.clear_screen()
                prev_term_dims = (vert_index, hor_index, term_width, term_height)
                render(vert_index, hor_index, term_width, term_height)
    except KeyboardInterrupt: ...
    finally:
        log_sched.shutdown()
        log_file.unlink(missing_ok=True)
        menus.clear_screen()
        menus.move_cursor((0, 0))
        menus._show()


# TODO(ayush): implement this
def _relaunch_modal(execution_data):
    return


def _abort_modal(execution_data):
    def render(term_width: int, term_height: int):
        # DISCLAIMER : MOST OF THE MAGIC NUMBERS HERE WERE THROUGH TRIAL AND ERROR
        menus.clear_screen()

        question = (
            f"Are you sure you want to abort {execution_data['display_name']} "
            f"({execution_data['workflow_name']}/{execution_data['workflow_version']})?"
        )
        lines = textwrap.wrap(question, width=term_width - 6)

        max_line_length = max(map(len, lines))

        x = (term_width - max_line_length) // 2
        y = (term_height - len(lines)) // 2

        menus.draw_box((x - 3, y - 2), len(lines) + 4, max_line_length + 4)

        for i, line in enumerate(lines):
            x = (term_width - len(line)) // 2
            menus.move_cursor((x, y + i))
            menus._print(f"{line}")

        ctrl_str = "[Y] Yes\t[N] No"
        menus.move_cursor(((term_width - len(ctrl_str)) // 2, y + len(lines) + 1))
        menus._print(ctrl_str)

    try:
        term_width, term_height = os.get_terminal_size()
        render(term_width, term_height)
        prev_term_dims = (term_width, term_height)
        while True:
            b = menus.read_bytes(1)
            if b in (b"y", b"Y"):
                headers = {"Authorization": f"Bearer {retrieve_or_login()}"}
                resp = post(
                    url=config.api.execution.abort,
                    headers=headers,
                    json={"execution_id": execution_data["id"]},
                )
                return
            elif b in (b"n", b"N"):
                return
            term_width, term_height = os.get_terminal_size()
            if prev_term_dims != (term_width, term_height):
                prev_term_dims = (term_width, term_height)
                render(term_width, term_height)
    except KeyboardInterrupt: ...
    finally:
        menus.clear_screen()
        menus.move_cursor((0, 0))



from urllib.parse import urljoin

import click
from latch_sdk_config.latch import NUCLEUS_URL

from .. import tinyrequests
from ..utils import get_auth_header


def stop_pod(pod_id: int) -> None:
    """Stops a pod given a pod_id"""
    res = tinyrequests.post(
        urljoin(NUCLEUS_URL, "/pods/stop"),
        headers={"Authorization": get_auth_header()},
        json={"pod_id": pod_id},
    )

    if res.status_code == 200:
        click.secho(f"Pod with ID `{pod_id}` stopped.", fg="green")
        return

    if res.status_code == 403 or res.status_code == 404:
        click.secho("Pod does not exist or permission denied.", fg="red")
        return

    if res.status_code != 200:
        click.secho(
            f"Internal error while stopping Pod `{pod_id}`. Please try again."
            " contact `support@latch.bio` if the issue persists.",
            fg="red",
        )
        return



from dataclasses import asdict, fields, is_dataclass, make_dataclass
from enum import Enum
from textwrap import dedent
from typing import Any, Dict, List, Optional, Type, Union, get_args, get_origin

import click
from flytekit.core.annotation import FlyteAnnotation
from typing_extensions import Annotated, TypeAlias, TypeGuard, TypeVar

from latch.types.directory import LatchDir
from latch.types.file import LatchFile
from latch_cli.utils import identifier_from_str

JSONValue: TypeAlias = Union[int, str, bool, float, None, List["JSONValue"], "JSONDict"]
JSONDict: TypeAlias = Dict[str, "JSONValue"]

# ayush: yoinked from console
valid_extensions = {
    "bed",
    "vcf",
    "css",
    "csv",
    "gif",
    "png",
    "pdf",
    "webp",
    "xhtml",
    "xlsx",
    "xml",
    "py",
    "log",
    "json",
    "gz",
    "mmtf",
    "deseqreport",
    "sam",
    "bam",
    "cram",
    "tsv",
    "tab",
    "sf",
    "txt",
    "text",
    "license",
    "readme",
    "r",
    "rscript",
    "md",
    "markdown",
    "markdn",
    "mdown",
    "htm",
    "html",
    "ipynb",
    "jpeg",
    "jpg",
    "jif",
    "jpe",
    "jfif",
    "js",
    "mjs",
    "es",
    "ts",
    "jsx",
    "tsx",
    "svg",
    "svgz",
    "fasta",
    "fna",
    "fa",
    "ffn",
    "faa",
    "frn",
    "fastq",
    "fq",
    "pdb",
    "pdb1",
    "ent",
    "brk",
    "ml2",
    "mol2",
    "sy2",
    "hdf",
    "h4",
    "hdf4",
    "he4",
    "h5",
    "hdf5",
    "he5",
    "h5ad",
    "yaml",
    "yml",
}


def parse_type(
    v: JSONValue, name: Optional[str] = None, *, infer_files: bool = False
) -> Type:
    if v is None:
        return str

    if infer_files and isinstance(v, str):
        if any([v.endswith(ext) for ext in valid_extensions]):
            return LatchFile
        elif v.endswith("/"):
            return LatchDir

    if is_primitive_value(v):
        return type(v)

    if isinstance(v, list):
        parsed_types = tuple(
            parse_type(
                x,
                name,
                infer_files=infer_files,
            )
            for x in v
        )

        if len(set(parsed_types)) != 1:
            raise ValueError(
                "Generic Lists are not supported - please"
                f" ensure that all elements in {name} are of the same type",
            )
        typ = parsed_types[0]
        if typ in {LatchFile, LatchDir}:
            return Annotated[List[typ], FlyteAnnotation({"size": len(v)})]
        return List[typ]

    assert isinstance(v, dict)

    if name is None:
        name = "SnakemakeRecord"

    fields: Dict[str, Type] = {}
    for k, x in v.items():
        fields[identifier_from_str(k)] = parse_type(
            x,
            k,
            infer_files=infer_files,
        )

    return make_dataclass(identifier_from_str(name), fields.items())


# returns raw value and generated default
def parse_value(t: Type, v: JSONValue):
    if v is None:
        return None, None

    if get_origin(t) is Annotated:
        args = get_args(t)
        assert len(args) > 0
        return parse_value(args[0], v)

    if t in {LatchFile, LatchDir}:
        # ayush: autogenerated defaults don't make sense for files/dirs since their
        # value in the config is their local path
        return v, None

    if is_primitive_value(v):
        return v, v

    if isinstance(v, list):
        assert get_origin(t) is list

        args = get_args(t)
        assert len(args) > 0

        sub_type = args[0]
        res = [parse_value(sub_type, x) for x in v]
        return [x[0] for x in res], [x[1] for x in res]

    assert isinstance(v, dict), v
    assert is_dataclass(t), t

    ret = {}
    defaults = {}
    fs = {identifier_from_str(f.name): f for f in fields(t)}

    for k, x in v.items():
        sanitized = identifier_from_str(k)
        assert sanitized in fs, sanitized
        val, default = parse_value(fs[sanitized].type, x)
        ret[sanitized] = val
        defaults[sanitized] = default

    return t(**ret), t(**defaults)


def is_primitive_type(
    typ: Type,
) -> TypeGuard[Union[Type[None], Type[str], Type[bool], Type[int], Type[float]]]:
    return typ in {Type[None], str, bool, int, float}


def is_primitive_value(val: object) -> TypeGuard[Union[None, str, bool, int, float]]:
    return is_primitive_type(type(val))


def is_list_type(typ: Type) -> TypeGuard[Type[List]]:
    return get_origin(typ) is list


def type_repr(t: Type, *, add_namespace: bool = False) -> str:
    if is_primitive_type(t) or t in {LatchFile, LatchDir}:
        return t.__name__

    if get_origin(t) is None:
        return f"{'latch_metadata.' if add_namespace else ''}{t.__name__}"

    if get_origin(t) is list:
        args = get_args(t)
        if len(args) > 0:
            return f"typing.List[{type_repr(args[0], add_namespace=add_namespace)}]"

        return "typing.List"

    if get_origin(t) is Union:
        args = get_args(t)

        if len(args) != 2 or args[1] is not type(None):
            raise ValueError(
                "Union types other than Optional are not yet supported in Snakemake"
                " workflows."
            )

        return f"typing.Optional[{type_repr(args[0], add_namespace=add_namespace)}]"

    if get_origin(t) is Annotated:
        args = get_args(t)
        assert len(args) > 1
        assert isinstance(args[1], FlyteAnnotation)
        return (
            f"typing_extensions.Annotated[{type_repr(args[0], add_namespace=add_namespace)},"
            f" FlyteAnnotation({repr(args[1].data)})]"
        )

    return t.__name__


def dataclass_repr(typ: Type) -> str:
    assert is_dataclass(typ)

    lines = ["@dataclass", f"class {typ.__name__}:"]
    for f in fields(typ):
        lines.append(f"    {f.name}: {type_repr(f.type)}")

    return "\n".join(lines) + "\n\n\n"


def get_preamble(typ: Type) -> str:
    if get_origin(typ) is Annotated:
        args = get_args(typ)
        assert len(args) > 0
        return get_preamble(args[0])

    if is_primitive_type(typ) or typ in {LatchFile, LatchDir}:
        return ""

    if get_origin(typ) in {Union, list}:
        return "".join([get_preamble(t) for t in get_args(typ)])

    assert is_dataclass(typ), typ

    preamble = "".join([get_preamble(f.type) for f in fields(typ)])

    return "".join([preamble, dataclass_repr(typ)])


def validate_snakemake_type(name: str, t: Type, param: Any) -> None:
    if t is type(None):
        return param is None

    elif is_primitive_type(t) or t in {LatchFile, LatchDir}:
        if param is None:
            raise ValueError(
                f"Parameter {name} of type {t} cannot be None. Either specify a"
                " non-None default value or use the Optional type"
            )
        if not isinstance(param, t):
            raise ValueError(f"Parameter {name} must be of type {t}, not {type(param)}")

    elif get_origin(t) is Union:
        args = get_args(t)
        # only Optional types supported
        if len(args) != 2 or args[1] is not type(None):
            raise ValueError(
                f"Failed to parse input param {param}. Union types other than"
                " Optional are not yet supported in Snakemake workflows."
            )
        if param is None:
            return
        validate_snakemake_type(name, args[0], param)

    elif get_origin(t) is Annotated:
        args = get_args(t)
        assert len(args) > 0
        validate_snakemake_type(name, args[0], param)

    elif is_list_type(t):
        args = get_args(t)
        if len(args) == 0:
            raise ValueError(
                "Generic Lists are not supported - please specify a subtype,"
                " e.g. List[LatchFile]",
            )
        list_typ = args[0]
        for i, val in enumerate(param):
            validate_snakemake_type(f"{name}[{i}]", list_typ, val)

    else:
        assert is_dataclass(t)
        for field in fields(t):
            validate_snakemake_type(
                f"{name}.{field.name}", field.type, getattr(param, field.name)
            )



from dataclasses import fields, is_dataclass
from pathlib import Path
from typing import Dict, List, Tuple, Type, TypeVar, get_args, get_origin

import click
import yaml
from typing_extensions import Annotated

from latch.types.directory import LatchDir
from latch.types.file import LatchFile
from latch_cli.snakemake.utils import reindent
from latch_cli.utils import best_effort_display_name, identifier_from_str

from .utils import (
    JSONValue,
    get_preamble,
    is_list_type,
    is_primitive_type,
    parse_type,
    parse_value,
    type_repr,
)

T = TypeVar("T")


def parse_config(
    config_path: Path,
    *,
    infer_files: bool = False,
) -> Dict[str, Tuple[Type[T], T]]:
    if not config_path.exists():
        click.secho(
            f"No config file found at {config_path}.",
            fg="red",
        )
        raise click.exceptions.Exit(1)

    if config_path.is_dir():
        click.secho(
            f"Path {config_path} points to a directory.",
            fg="red",
        )
        raise click.exceptions.Exit(1)

    try:
        res: JSONValue = yaml.safe_load(config_path.read_text())
    except yaml.YAMLError as e:
        click.secho(
            reindent(
                f"""
                Error loading config from {config_path}:

                {e}
                """,
                0,
            ),
            fg="red",
        )
        raise click.exceptions.Exit(1) from e

    if not isinstance(res, dict):
        # ayush: this case doesn't matter bc a non-dict .yaml file isn't valid snakemake
        return {"snakemake_parameter": (parse_type(res, infer_files=infer_files), res)}

    parsed: Dict[str, Type] = {}
    for k, v in res.items():
        try:
            typ = parse_type(v, k, infer_files=infer_files)
        except ValueError as e:
            click.secho(
                f"WARNING: Skipping parameter {k}. Failed to parse type: {e}.",
                fg="yellow",
            )
            continue
        val, default = parse_value(typ, v)

        parsed[k] = (typ, (val, default))

    return parsed


def file_metadata_str(typ: Type, value: JSONValue, level: int = 0) -> str:
    if get_origin(typ) is Annotated:
        args = get_args(typ)
        assert len(args) > 0
        return file_metadata_str(args[0], value, level)

    if is_primitive_type(typ):
        return ""

    if typ in {LatchFile, LatchDir}:
        return reindent(
            f"""\
            SnakemakeFileMetadata(
                path={repr(value)},
                config=True,
            ),\n""",
            level,
        )

    metadata: List[str] = []
    if is_list_type(typ):
        template = """
        [
        __metadata__],\n"""

        args = get_args(typ)
        assert len(args) > 0
        for val in value:
            metadata_str = file_metadata_str(get_args(typ)[0], val, level + 1)
            if metadata_str == "":
                continue
            metadata.append(metadata_str)
    else:
        template = """
        {
        __metadata__},\n"""

        assert is_dataclass(typ)
        for field in fields(typ):
            metadata_str = file_metadata_str(
                field.type, getattr(value, field.name), level
            )
            if metadata_str == "":
                continue
            metadata_str = f"{repr(identifier_from_str(field.name))}: {metadata_str}"
            metadata.append(reindent(metadata_str, level + 1))

    if len(metadata) == 0:
        return ""

    return reindent(
        template,
        level,
    ).replace("__metadata__", "".join(metadata), level + 1)


# todo(ayush): print informative stuff here ala register
def generate_metadata(
    config_path: Path,
    *,
    skip_confirmation: bool = False,
    generate_defaults: bool = False,
    infer_files: bool = False,
):
    parsed = parse_config(config_path, infer_files=infer_files)

    preambles: List[str] = []
    params: List[str] = []
    file_metadata: List[str] = []

    for k, (typ, (val, default)) in parsed.items():
        preambles.append(get_preamble(typ))

        param_str = reindent(
            f"""\
            {repr(identifier_from_str(k))}: SnakemakeParameter(
                display_name={repr(best_effort_display_name(k))},
                type={type_repr(typ)},
            __default__),""",
            0,
        )

        default_str = ""
        if generate_defaults and default is not None:
            default_str = f"    default={repr(default)},\n"

        param_str = param_str.replace("__default__", default_str)

        param_str = reindent(param_str, 1)
        params.append(param_str)

        metadata_str = file_metadata_str(typ, val)
        if metadata_str == "":
            continue
        metadata_str = f"{repr(identifier_from_str(k))}: {metadata_str}"
        file_metadata.append(reindent(metadata_str, 1))

    metadata_root = Path("latch_metadata")
    if metadata_root.is_file():
        if not click.confirm("A file exists at `latch_metadata`. Delete it?"):
            raise click.exceptions.Exit(0)

        metadata_root.unlink()

    metadata_root.mkdir(exist_ok=True)

    metadata_path = metadata_root / Path("__init__.py")
    old_metadata_path = Path("latch_metadata.py")

    if old_metadata_path.exists() and not metadata_path.exists():
        if click.confirm(
            "Found legacy `latch_metadata.py` file in current directory. This is"
            " deprecated and will be ignored in future releases. Move to"
            " `latch_metadata/__init__.py`? (This will not change file contents)"
        ):
            old_metadata_path.rename(metadata_path)
    elif old_metadata_path.exists() and metadata_path.exists():
        click.secho(
            "Warning: Found both `latch_metadata.py` and"
            " `latch_metadata/__init__.py` in current directory."
            " `latch_metadata.py` will be ignored.",
            fg="yellow",
        )

    if not metadata_path.exists():
        metadata_path.write_text(
            reindent(
                r"""
                from latch.types.metadata import SnakemakeMetadata, LatchAuthor, EnvironmentConfig
                from latch.types.directory import LatchDir

                from .parameters import generated_parameters, file_metadata

                SnakemakeMetadata(
                    output_dir=LatchDir("latch:///your_output_directory"),
                    display_name="Your Workflow Name",
                    author=LatchAuthor(
                        name="Your Name",
                    ),
                    env_config=EnvironmentConfig(
                        use_conda=False,
                        use_container=False,
                    ),
                    cores=4,
                    # Add more parameters
                    parameters=generated_parameters,
                    file_metadata=file_metadata,

                )
                """,
                0,
            )
        )
        click.secho("Generated `latch_metadata/__init__.py`.", fg="green")

    params_path = metadata_root / Path("parameters.py")
    if (
        params_path.exists()
        and not skip_confirmation
        and not click.confirm(
            "File `latch_metadata/parameters.py` already exists. Overwrite?"
        )
    ):
        raise click.exceptions.Exit(0)

    params_path.write_text(
        reindent(
            r"""
            from dataclasses import dataclass
            import typing
            import typing_extensions

            from flytekit.core.annotation import FlyteAnnotation

            from latch.types.metadata import SnakemakeParameter, SnakemakeFileParameter, SnakemakeFileMetadata
            from latch.types.file import LatchFile
            from latch.types.directory import LatchDir

            __preambles__

            # Import these into your `__init__.py` file:
            #
            # from .parameters import generated_parameters, file_metadata

            generated_parameters = {
            __params__
            }

            file_metadata = {
            __file_metadata__}

            """,
            0,
        )
        .replace("__preambles__", "".join(preambles))
        .replace("__params__", "\n".join(params))
        .replace("__file_metadata__", "".join(file_metadata))
    )
    click.secho("Generated `latch_metadata/parameters.py`.", fg="green")
